from infrahub.core.validators.attribute.enum import AttributeEnumChecker
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class SchemaAttributeEnumConstraintDependency(DependencyBuilder[AttributeEnumChecker]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> AttributeEnumChecker:
        return AttributeEnumChecker(db=context.db, branch=context.branch)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING

from infrahub.core.account import GlobalPermission
from infrahub.core.constants import GlobalPermissions, PermissionDecision
from infrahub.exceptions import ValidationError

from .checker import verify_proposed_change_is_mergeable
from .constants import ProposedChangeAction, ProposedChangeState

if TYPE_CHECKING:
    from collections.abc import Sequence

    from infrahub.core.protocols import CoreGenericAccount, CoreProposedChange
    from infrahub.graphql.initialization import GraphqlContext


class Check(ABC):
    @abstractmethod
    async def evaluate(
        self,
        proposed_change: CoreProposedChange,
        proposed_change_author: CoreGenericAccount,
        graphql_context: GraphqlContext,
    ) -> None: ...


class IsAuthor(Check):
    async def evaluate(
        self,
        proposed_change: CoreProposedChange,  # noqa: ARG002
        proposed_change_author: CoreGenericAccount,
        graphql_context: GraphqlContext,
    ) -> None:
        if proposed_change_author.id != graphql_context.active_account_session.account_id:
            raise ValidationError("You are not the author of the proposed change")


class StateIs(Check):
    def __init__(self, expected: Sequence[ProposedChangeState]) -> None:
        self.expected = expected

    async def evaluate(
        self,
        proposed_change: CoreProposedChange,
        proposed_change_author: CoreGenericAccount,  # noqa: ARG002
        graphql_context: GraphqlContext,  # noqa: ARG002
    ) -> None:
        if proposed_change.state.value.value not in self.expected:
            raise ValidationError(f"The proposed change is not {', '.join([i.value for i in self.expected])}")


class DraftIs(Check):
    def __init__(self, expected: bool) -> None:
        self.expected = expected

    async def evaluate(
        self,
        proposed_change: CoreProposedChange,
        proposed_change_author: CoreGenericAccount,  # noqa: ARG002
        graphql_context: GraphqlContext,  # noqa: ARG002
    ) -> None:
        if proposed_change.is_draft.value != self.expected:
            if self.expected:
                raise ValidationError("The proposed change is not a draft")
            raise ValidationError("The proposed change is a draft")


class HasPermission(Check):
    def __init__(self, permission: GlobalPermission) -> None:
        self.permission = permission

    async def evaluate(
        self,
        proposed_change: CoreProposedChange,  # noqa: ARG002
        proposed_change_author: CoreGenericAccount,  # noqa: ARG002
        graphql_context: GraphqlContext,
    ) -> None:
        if not graphql_context.active_permissions.has_permission(permission=self.permission):
            raise ValidationError("You do not have the permission to perform this action")


class IsMergeable(Check):
    async def evaluate(
        self,
        proposed_change: CoreProposedChange,
        proposed_change_author: CoreGenericAccount,  # noqa: ARG002
        graphql_context: GraphqlContext,
    ) -> None:
        try:
            await verify_proposed_change_is_mergeable(
                proposed_change=proposed_change,  # type: ignore[arg-type]
                db=graphql_context.db,
                account_session=graphql_context.active_account_session,
            )
        except ValueError as exc:
            raise ValidationError(str(exc)) from exc


@dataclass
class ActionRule:
    action: ProposedChangeAction
    checks: list[Check]

    async def evaluate(
        self,
        proposed_change: CoreProposedChange,
        proposed_change_author: CoreGenericAccount,
        graphql_context: GraphqlContext,
    ) -> dict[str, str | bool | None]:
        for check in self.checks:
            try:
                await check.evaluate(
                    proposed_change=proposed_change,
                    proposed_change_author=proposed_change_author,
                    graphql_context=graphql_context,
                )
            except ValidationError as exc:
                return {"action": self.action.value, "available": False, "unavailability_reason": exc.message}

        return {"action": self.action.value, "available": True, "unavailability_reason": None}


class ActionRulesEvaluator:
    def __init__(self, rules: list[ActionRule]):
        self.rules = rules

    async def evaluate(
        self,
        proposed_change: CoreProposedChange,
        proposed_change_author: CoreGenericAccount,
        graphql_context: GraphqlContext,
    ) -> list[dict[str, str | bool | None]]:
        report: list[dict[str, str | bool | None]] = []
        for rule in self.rules:
            report.append(
                await rule.evaluate(
                    proposed_change=proposed_change,
                    proposed_change_author=proposed_change_author,
                    graphql_context=graphql_context,
                )
            )
        return report


MERGE_PROPOSED_CHANGE_PERMISSION = GlobalPermission(
    action=GlobalPermissions.MERGE_PROPOSED_CHANGE.value,
    decision=PermissionDecision.ALLOW_ALL.value,
)
REVIEW_PROPOSED_CHANGE_PERMISSION = GlobalPermission(
    action=GlobalPermissions.REVIEW_PROPOSED_CHANGE.value,
    decision=PermissionDecision.ALLOW_ALL.value,
)

ACTION_RULES = [
    ActionRule(action=ProposedChangeAction.OPEN, checks=[StateIs(expected=[ProposedChangeState.CLOSED])]),
    ActionRule(action=ProposedChangeAction.CLOSE, checks=[StateIs(expected=[ProposedChangeState.OPEN])]),
    ActionRule(
        action=ProposedChangeAction.SET_DRAFT,
        checks=[IsAuthor(), StateIs(expected=[ProposedChangeState.OPEN]), DraftIs(expected=False)],
    ),
    ActionRule(
        action=ProposedChangeAction.UNSET_DRAFT,
        checks=[IsAuthor(), StateIs(expected=[ProposedChangeState.OPEN]), DraftIs(expected=True)],
    ),
    ActionRule(
        action=ProposedChangeAction.APPROVE,
        checks=[
            StateIs(expected=[ProposedChangeState.OPEN]),
            HasPermission(permission=REVIEW_PROPOSED_CHANGE_PERMISSION),
        ],
    ),
    ActionRule(
        action=ProposedChangeAction.CANCEL_APPROVE,
        checks=[
            StateIs(expected=[ProposedChangeState.OPEN]),
            HasPermission(permission=REVIEW_PROPOSED_CHANGE_PERMISSION),
        ],
    ),
    ActionRule(
        action=ProposedChangeAction.REJECT,
        checks=[
            StateIs(expected=[ProposedChangeState.OPEN]),
            HasPermission(permission=REVIEW_PROPOSED_CHANGE_PERMISSION),
        ],
    ),
    ActionRule(
        action=ProposedChangeAction.CANCEL_REJECT,
        checks=[
            StateIs(expected=[ProposedChangeState.OPEN]),
            HasPermission(permission=REVIEW_PROPOSED_CHANGE_PERMISSION),
        ],
    ),
    ActionRule(
        action=ProposedChangeAction.MERGE,
        checks=[
            StateIs(expected=[ProposedChangeState.OPEN]),
            DraftIs(expected=False),
            HasPermission(permission=MERGE_PROPOSED_CHANGE_PERMISSION),
            IsMergeable(),
        ],
    ),
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.constants import NULL_VALUE, PathType
from infrahub.core.path import DataPath, GroupedDataPaths

from ..interface import ConstraintCheckerInterface
from ..shared import AttributeSchemaValidatorQuery

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase

    from ..model import SchemaConstraintValidatorRequest


class AttributeChoicesUpdateValidatorQuery(AttributeSchemaValidatorQuery):
    name: str = "attribute_constraints_choices_validator"

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        if self.attribute_schema.choices is None:
            return

        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string())
        self.params.update(branch_params)

        self.params["attr_name"] = self.attribute_schema.name
        self.params["allowed_values"] = [choice.name for choice in self.attribute_schema.choices]
        self.params["null_value"] = NULL_VALUE

        query = """
        MATCH p = (n:%(node_kind)s)
        CALL (n) {
            MATCH path = (root:Root)<-[rr:IS_PART_OF]-(n)-[ra:HAS_ATTRIBUTE]-(:Attribute { name: $attr_name } )-[rv:HAS_VALUE]-(av:AttributeValue)
            WHERE all(
                r in relationships(path)
                WHERE %(branch_filter)s
            )
            RETURN path as full_path, n as node, rv as value_relationship, av.value as attribute_value
            ORDER BY rv.branch_level DESC, ra.branch_level DESC, rr.branch_level DESC, rv.from DESC, ra.from DESC, rr.from DESC
            LIMIT 1
        }
        WITH full_path, node, attribute_value, value_relationship
        WHERE all(r in relationships(full_path) WHERE r.status = "active")
        AND attribute_value IS NOT NULL
        AND attribute_value <> $null_value
        AND NOT (attribute_value IN $allowed_values)
        """ % {"branch_filter": branch_filter, "node_kind": self.node_schema.kind}

        self.add_to_query(query)
        self.return_labels = ["node.uuid", "attribute_value", "value_relationship"]

    async def get_paths(self) -> GroupedDataPaths:
        grouped_data_paths = GroupedDataPaths()
        for result in self.results:
            value = str(result.get("attribute_value"))
            grouped_data_paths.add_data_path(
                DataPath(
                    branch=str(result.get("value_relationship").get("branch")),
                    path_type=PathType.ATTRIBUTE,
                    node_id=str(result.get("node.uuid")),
                    field_name=self.attribute_schema.name,
                    kind=self.node_schema.kind,
                    value=value,
                ),
                grouping_key=value,
            )

        return grouped_data_paths


class AttributeChoicesChecker(ConstraintCheckerInterface):
    query_classes = [AttributeChoicesUpdateValidatorQuery]

    def __init__(self, db: InfrahubDatabase, branch: Branch | None = None):
        self.db = db
        self.branch = branch

    @property
    def name(self) -> str:
        return "attribute.choices.update"

    def supports(self, request: SchemaConstraintValidatorRequest) -> bool:
        return request.constraint_name == self.name

    async def check(self, request: SchemaConstraintValidatorRequest) -> list[GroupedDataPaths]:
        grouped_data_paths_list: list[GroupedDataPaths] = []
        if not request.schema_path.field_name:
            raise ValueError("field_name is not defined")
        attribute_schema = request.node_schema.get_attribute(name=request.schema_path.field_name)
        if attribute_schema.choices is None:
            return grouped_data_paths_list

        for query_class in self.query_classes:
            # TODO add exception handling
            query = await query_class.init(
                db=self.db, branch=self.branch, node_schema=request.node_schema, schema_path=request.schema_path
            )
            await query.execute(db=self.db)
            grouped_data_paths_list.append(await query.get_paths())
        return grouped_data_paths_list



import uuid

from pydantic import BaseModel, ConfigDict, Field

from infrahub.core.constants import CheckType
from infrahub.generators.models import ProposedChangeGeneratorDefinition
from infrahub.message_bus import InfrahubMessage
from infrahub.message_bus.types import ProposedChangeArtifactDefinition, ProposedChangeBranchDiff


class BaseProposedChangeWithDiffMessage(InfrahubMessage):
    """Sent trigger the refresh of artifacts that are impacted by the proposed change."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    proposed_change: str = Field(..., description="The unique ID of the Proposed Change")
    source_branch: str = Field(..., description="The source branch of the proposed change")
    source_branch_sync_with_git: bool = Field(..., description="Indicates if the source branch should sync with git")
    destination_branch: str = Field(..., description="The destination branch of the proposed change")
    branch_diff: ProposedChangeBranchDiff = Field(..., description="The calculated diff between the two branches")


class RequestProposedChangeDataIntegrity(BaseProposedChangeWithDiffMessage):
    """Sent trigger data integrity checks for a proposed change"""


class RequestProposedChangeRunGenerators(BaseProposedChangeWithDiffMessage):
    """Sent trigger the generators that are impacted by the proposed change to run."""

    refresh_artifacts: bool = Field(..., description="Whether to regenerate artifacts after the generators are run")
    do_repository_checks: bool = Field(
        ..., description="Whether to run repository and user checks after the generators are run"
    )


class RequestProposedChangeRepositoryChecks(BaseProposedChangeWithDiffMessage):
    """Sent when a proposed change is created to trigger additional checks"""


class RequestProposedChangeSchemaIntegrity(BaseProposedChangeWithDiffMessage):
    """Sent trigger schema integrity checks for a proposed change"""


class RequestProposedChangeUserTests(BaseProposedChangeWithDiffMessage):
    """Sent trigger to run tests (smoke, units, integrations) for a proposed change."""


class RequestArtifactDefinitionCheck(BaseModel):
    """Sent to validate the generation of artifacts in relation to a proposed change."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    artifact_definition: ProposedChangeArtifactDefinition = Field(..., description="The Artifact Definition")
    branch_diff: ProposedChangeBranchDiff = Field(..., description="The calculated diff between the two branches")
    proposed_change: str = Field(..., description="The unique ID of the Proposed Change")
    source_branch: str = Field(..., description="The source branch")
    source_branch_sync_with_git: bool = Field(..., description="Indicates if the source branch should sync with git")
    destination_branch: str = Field(..., description="The target branch")


class RunGeneratorAsCheckModel(BaseModel):
    """A check that runs a generator."""

    generator_definition: ProposedChangeGeneratorDefinition = Field(..., description="The Generator definition")
    generator_instance: str | None = Field(
        default=None, description="The id of the generator instance if it previously existed"
    )
    commit: str = Field(..., description="The commit to target")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the Repository")
    repository_kind: str = Field(..., description="The kind of the Repository")
    branch_name: str = Field(..., description="The branch where the check is run")
    target_id: str = Field(..., description="The ID of the target object for this generator")
    target_name: str = Field(..., description="Name of the generator target")
    query: str = Field(..., description="The name of the query to use when collecting data")
    variables: dict = Field(..., description="Input variables when running the generator")
    validator_id: str = Field(..., description="The ID of the validator")
    proposed_change: str = Field(..., description="The unique ID of the Proposed Change")


class RequestGeneratorDefinitionCheck(BaseModel):
    """Sent to trigger Generators to run for a proposed change."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    generator_definition: ProposedChangeGeneratorDefinition = Field(..., description="The Generator Definition")
    branch_diff: ProposedChangeBranchDiff = Field(..., description="The calculated diff between the two branches")
    proposed_change: str = Field(..., description="The unique ID of the Proposed Change")
    source_branch: str = Field(..., description="The source branch")
    source_branch_sync_with_git: bool = Field(..., description="Indicates if the source branch should sync with git")
    destination_branch: str = Field(..., description="The target branch")


class RequestProposedChangePipeline(BaseModel):
    """Sent request the start of a pipeline connected to a proposed change."""

    proposed_change: str = Field(..., description="The unique ID of the proposed change")
    source_branch: str = Field(..., description="The source branch of the proposed change")
    source_branch_sync_with_git: bool = Field(..., description="Indicates if the source branch should sync with git")
    destination_branch: str = Field(..., description="The destination branch of the proposed change")
    check_type: CheckType = Field(
        default=CheckType.ALL, description="Can be used to restrict the pipeline to a specific type of job"
    )
    pipeline_id: uuid.UUID = Field(
        default_factory=uuid.uuid4, description="The unique ID of the execution of this pipeline"
    )


class RequestProposedChangeRefreshArtifacts(BaseProposedChangeWithDiffMessage):
    """Sent trigger the refresh of artifacts that are impacted by the proposed change."""



from typing import Any

from pydantic import BaseModel, ConfigDict
from typing_extensions import Self

from infrahub.core.query import Query, QueryResult, QueryType
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase

from ..model.path import TrackingId
from .diff_get import QUERY_MATCH_NODES
from .filters import EnrichedDiffQueryFilters


# type: ignore[call-overload]
class DiffSummaryCounters(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    num_added: int = 0
    num_updated: int = 0
    num_unchanged: int = 0
    num_removed: int = 0
    num_conflicts: int = 0
    from_time: Timestamp
    to_time: Timestamp

    @classmethod
    def from_graph(cls, result: QueryResult) -> Self:
        return cls(
            num_added=int(result.get_as_str("num_added") or 0),
            num_updated=int(result.get_as_str("num_updated") or 0),
            num_unchanged=int(result.get_as_str("num_unchanged") or 0),
            num_removed=int(result.get_as_str("num_removed") or 0),
            num_conflicts=int(result.get_as_str("num_conflicts") or 0),
            from_time=Timestamp(result.get_as_str("from_time")),
            to_time=Timestamp(result.get_as_str("to_time")),
        )


class DiffSummaryQuery(Query):
    """Get a Summary of the diff"""

    name = "enriched_diff_summary"
    type = QueryType.READ
    insert_limit = False

    def __init__(
        self,
        base_branch_name: str,
        diff_branch_names: list[str],
        filters: EnrichedDiffQueryFilters,
        from_time: Timestamp | None = None,
        to_time: Timestamp | None = None,
        tracking_id: TrackingId | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.base_branch_name = base_branch_name
        self.diff_branch_names = diff_branch_names
        self.filters = filters or EnrichedDiffQueryFilters()
        self.from_time = from_time
        self.to_time = to_time
        self.tracking_id = tracking_id

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        if (self.from_time is None or self.to_time is None) and self.tracking_id is None:
            raise ValueError("DiffSummaryQuery requires from_time and to_time or tracking_id ")
        self.params = {
            "base_branch": self.base_branch_name,
            "diff_branches": self.diff_branch_names,
            "from_time": self.from_time.to_string() if self.from_time else None,
            "to_time": self.to_time.to_string() if self.to_time else None,
            "tracking_id": self.tracking_id.serialize() if self.tracking_id else None,
            "diff_ids": None,
        }

        self.add_to_query(query=QUERY_MATCH_NODES)

        if not self.filters.is_empty:
            filters, filter_params = self.filters.generate()
            self.params.update(filter_params)

            query_filters = """
            WHERE (
                %(filters)s
            )
            """ % {"filters": filters}
            self.add_to_query(query=query_filters)

        self.return_labels = [
            "SUM(diff_node.num_added) as num_added",
            "SUM(diff_node.num_updated) as num_updated",
            "SUM(diff_node.num_unchanged) as num_unchanged",
            "SUM(diff_node.num_removed) as num_removed",
            "SUM(diff_node.num_conflicts) as num_conflicts",
            "min(diff_root.from_time) as from_time",
            "max(diff_root.to_time) as to_time",
            "count(diff_root) as num_roots",
        ]

    def get_summary(self) -> DiffSummaryCounters | None:
        result = self.get_result()
        if not result:
            return None
        if result.get("num_roots") == 0:
            return None
        return DiffSummaryCounters.from_graph(result)



from __future__ import annotations

from pydantic import Field

from infrahub.message_bus import InfrahubMessage, InfrahubResponse, InfrahubResponseData

ROUTING_KEY = "git.repository.connectivity"


class GitRepositoryConnectivity(InfrahubMessage):
    """Validate connectivity and credentials to remote repository"""

    repository_name: str = Field(..., description="The name of the repository")
    repository_location: str = Field(..., description="The location of repository")


class GitRepositoryConnectivityResponseData(InfrahubResponseData):
    message: str = Field(..., description="The status message")
    success: bool = Field(...)


class GitRepositoryConnectivityResponse(InfrahubResponse):
    routing_key: str = ROUTING_KEY
    data: GitRepositoryConnectivityResponseData = Field(default=...)



import bcrypt


def hash_password(password: str) -> str:
    """Hash a password for storage in the database."""
    salt = bcrypt.gensalt()
    hashed_password = bcrypt.hashpw(password.encode("UTF-8"), salt)
    return hashed_password.decode("UTF-8")



from __future__ import annotations

from .models import Task, TaskFilter, TaskLog, TaskRelatedNode, TaskState

__all__ = [
    "Task",
    "TaskFilter",
    "TaskLog",
    "TaskRelatedNode",
    "TaskState",
]



from __future__ import annotations

import hashlib
from typing import TYPE_CHECKING

from fastapi import APIRouter, Depends, File, Response, UploadFile
from infrahub_sdk.uuidt import UUIDT
from pydantic import BaseModel

from infrahub.api.dependencies import get_current_user
from infrahub.core import registry
from infrahub.log import get_logger

if TYPE_CHECKING:
    from infrahub.auth import AccountSession

log = get_logger()
router = APIRouter(prefix="/storage")


class UploadResponse(BaseModel):
    identifier: str
    checksum: str


class UploadContentPayload(BaseModel):
    content: str


@router.get("/object/{identifier:str}")
def get_file(identifier: str, _: AccountSession = Depends(get_current_user)) -> Response:
    content = registry.storage.retrieve(identifier=identifier)
    return Response(content=content)


@router.post("/upload/content")
def upload_content(
    item: UploadContentPayload,
    _: str = Depends(get_current_user),
) -> UploadResponse:
    # TODO need to optimized how we read the content of the file, especially if the file is really large
    # Check this discussion for more details
    # https://stackoverflow.com/questions/63048825/how-to-upload-file-using-fastapi

    file_content = bytes(item.content, encoding="utf-8")
    identifier = str(UUIDT())

    checksum = hashlib.md5(file_content, usedforsecurity=False).hexdigest()
    registry.storage.store(identifier=identifier, content=file_content)
    return UploadResponse(identifier=identifier, checksum=checksum)


@router.post("/upload/file")
def upload_file(file: UploadFile = File(...), _: AccountSession = Depends(get_current_user)) -> UploadResponse:
    # TODO need to optimized how we read the content of the file, especially if the file is really large
    # Check this discussion for more details
    # https://stackoverflow.com/questions/63048825/how-to-upload-file-using-fastapi

    file_content = file.file.read()
    identifier = str(UUIDT())

    checksum = hashlib.md5(file_content, usedforsecurity=False).hexdigest()
    registry.storage.store(identifier=identifier, content=file_content)
    return UploadResponse(identifier=identifier, checksum=checksum)



from infrahub.core.validators.node.generate_profile import NodeGenerateProfileChecker
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class SchemaGenerateProfileConstraintDependency(DependencyBuilder[NodeGenerateProfileChecker]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> NodeGenerateProfileChecker:
        return NodeGenerateProfileChecker(db=context.db, branch=context.branch)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Self

from graphene import Boolean, Enum, Field, InputObjectType, List, Mutation, String
from graphql import GraphQLResolveInfo

from infrahub import lock
from infrahub.core.account import GlobalPermission
from infrahub.core.branch import Branch
from infrahub.core.constants import (
    CheckType,
    GlobalPermissions,
    InfrahubKind,
    PermissionDecision,
)
from infrahub.core.manager import NodeManager
from infrahub.core.protocols import CoreProposedChange
from infrahub.core.schema import NodeSchema
from infrahub.database import InfrahubDatabase, retry_db_transaction
from infrahub.events import (
    EventMeta,
    ProposedChangeApprovalRevokedEvent,
    ProposedChangeApprovedEvent,
    ProposedChangeRejectedEvent,
    ProposedChangeRejectionRevokedEvent,
)
from infrahub.exceptions import BranchNotFoundError, PermissionDeniedError, ValidationError
from infrahub.graphql.mutations.main import InfrahubMutationMixin
from infrahub.graphql.types.enums import CheckType as GraphQLCheckType
from infrahub.graphql.types.task import TaskInfo
from infrahub.lock import InfrahubLock, build_object_lock_name
from infrahub.proposed_change.approval_revoker import do_revoke_approvals_on_updated_pcs
from infrahub.proposed_change.constants import ProposedChangeApprovalDecision, ProposedChangeState
from infrahub.proposed_change.models import RequestProposedChangePipeline
from infrahub.workers.dependencies import get_event_service
from infrahub.workflows.catalogue import PROPOSED_CHANGE_MERGE, REQUEST_PROPOSED_CHANGE_PIPELINE

from .main import InfrahubMutationOptions

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.node import Node
    from infrahub.events.models import InfrahubEvent

    from ..initialization import GraphqlContext

ProposedChangeApprovalDecisionInput = Enum.from_enum(ProposedChangeApprovalDecision)


class InfrahubProposedChangeMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(
        cls, schema: NodeSchema, _meta: InfrahubMutationOptions | None = None, **options: dict[str, Any]
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)
        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    @retry_db_transaction(name="proposed_change_create")
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,  # noqa: ARG003
        override_data: dict[str, Any] | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context

        override_data = {"created_by": {"id": graphql_context.active_account_session.account_id}}
        state = data.get("state", {}).get("value")
        if state and state != ProposedChangeState.OPEN.value:
            raise ValidationError(input_value="A proposed change has to be in the open state during creation")

        async with graphql_context.db.start_transaction() as dbt:
            proposed_change, result = await super().mutate_create(
                info=info, data=data, branch=branch, database=dbt, override_data=override_data
            )
            destination_branch = proposed_change.destination_branch.value
            source_branch = await _get_source_branch(db=dbt, name=proposed_change.source_branch.value)
            if destination_branch == source_branch.name:
                raise ValidationError(input_value="The source and destination branch can't be the same")
            if destination_branch != "main":
                raise ValidationError(
                    input_value="Currently only the 'main' branch is supported as a destination for a proposed change"
                )

        if graphql_context.service:
            request_proposed_change_model = RequestProposedChangePipeline(
                proposed_change=proposed_change.id,
                source_branch=source_branch.name,
                source_branch_sync_with_git=source_branch.sync_with_git,
                destination_branch=destination_branch,
            )

            await graphql_context.service.workflow.submit_workflow(
                workflow=REQUEST_PROPOSED_CHANGE_PIPELINE,
                parameters={"model": request_proposed_change_model},
                context=graphql_context.get_context(),
            )

        return proposed_change, result

    @classmethod
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,  # noqa: ARG003
        node: Node | None = None,  # noqa: ARG003
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context

        obj = await NodeManager.get_one_by_id_or_default_filter(
            db=graphql_context.db,
            kind=cls._meta.schema.kind,
            id=data.get("id"),
            branch=branch,
            include_owner=True,
            include_source=True,
        )
        state = ProposedChangeState(obj.state.value.value)
        state.validate_updatable()

        updated_state = None
        if state_update := data.get("state", {}).get("value"):
            updated_state = ProposedChangeState(state_update)
            state.validate_state_transition(updated_state)

        # Check before starting a transaction, stopping in the middle of the transaction seems to break with memgraph
        if updated_state == ProposedChangeState.MERGED and graphql_context.account_session:
            try:
                graphql_context.active_permissions.raise_for_permission(
                    permission=GlobalPermission(
                        action=GlobalPermissions.MERGE_PROPOSED_CHANGE.value,
                        decision=PermissionDecision.ALLOW_ALL.value,
                    )
                )
            except PermissionDeniedError as exc:
                raise ValidationError(str(exc)) from exc

        if updated_state == ProposedChangeState.MERGED:
            data["state"]["value"] = ProposedChangeState.MERGING.value

        proposed_change, result = await super().mutate_update(
            info=info, data=data, branch=branch, database=graphql_context.db, node=obj
        )

        if updated_state == ProposedChangeState.MERGED:
            await graphql_context.service.workflow.execute_workflow(
                workflow=PROPOSED_CHANGE_MERGE,
                context=graphql_context.get_context(),
                parameters={
                    "proposed_change_id": proposed_change.id,
                    "proposed_change_name": proposed_change.name.value,
                },
            )
            # When the PROPOSED_CHANGE_MERGE succeeds it will have correctly changed the state
            # from the overridden "merging" value, so here we change it back to reflect the
            # correct value for the event that will be generated.
            proposed_change.node_changelog.attributes["state"].value = ProposedChangeState.MERGED.value

        return proposed_change, result


class ProposedChangeRequestRunCheckInput(InputObjectType):
    id = String(required=True)
    check_type = GraphQLCheckType(required=False)


class ProposedChangeRequestRunCheck(Mutation):
    class Arguments:
        data = ProposedChangeRequestRunCheckInput(required=True)

    ok = Boolean()

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: dict[str, Any],
    ) -> dict[str, bool]:
        graphql_context: GraphqlContext = info.context

        check_type = data.get("check_type") or CheckType.ALL

        identifier = data.get("id", "")
        proposed_change = await NodeManager.get_one_by_id_or_default_filter(
            id=identifier, kind=InfrahubKind.PROPOSEDCHANGE, db=graphql_context.db
        )
        state = ProposedChangeState(proposed_change.state.value.value)
        state.validate_state_check_run()

        destination_branch = proposed_change.destination_branch.value
        source_branch = await _get_source_branch(db=graphql_context.db, name=proposed_change.source_branch.value)

        request_proposed_change_model = RequestProposedChangePipeline(
            proposed_change=proposed_change.id,
            source_branch=source_branch.name,
            source_branch_sync_with_git=source_branch.sync_with_git,
            destination_branch=destination_branch,
            check_type=check_type,
        )
        if graphql_context.service:
            await graphql_context.service.workflow.submit_workflow(
                workflow=REQUEST_PROPOSED_CHANGE_PIPELINE,
                parameters={"model": request_proposed_change_model},
                context=graphql_context.get_context(),
            )

        return {"ok": True}


class ProposedChangeReviewInput(InputObjectType):
    id = String(required=True, description="The ID of the proposed change to review.")
    decision = ProposedChangeApprovalDecisionInput(
        required=True, description="The decision for the proposed change review."
    )


class ProposedChangeReview(Mutation):
    class Arguments:
        data = ProposedChangeReviewInput(required=True)

    ok = Boolean()

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: ProposedChangeReviewInput,
    ) -> dict[str, bool]:
        """
        This mutation is used to approve or reject a proposed change.
        It can also be used to undo an approval or rejection.
        """

        graphql_context: GraphqlContext = info.context
        graphql_context.active_permissions.raise_for_permission(
            permission=GlobalPermission(
                action=GlobalPermissions.REVIEW_PROPOSED_CHANGE.value, decision=PermissionDecision.ALLOW_ALL.value
            )
        )
        pc_id = str(data.id)
        lock_name = build_object_lock_name(pc_id)
        async with InfrahubLock(name=lock_name, connection=lock.registry.connection):
            proposed_change = await NodeManager.get_one_by_id_or_default_filter(
                id=pc_id, kind=CoreProposedChange, db=graphql_context.db, prefetch_relationships=True
            )
            state = ProposedChangeState(proposed_change.state.value.value)
            state.validate_reviewable()

            created_by = await proposed_change.created_by.get_peer(db=graphql_context.db)
            if created_by and created_by.id == graphql_context.active_account_session.account_id:
                raise ValidationError(input_value="You cannot review your own proposed changes")

            current_user = await NodeManager.get_one_by_id_or_default_filter(
                id=graphql_context.active_account_session.account_id,
                kind=InfrahubKind.GENERICACCOUNT,
                db=graphql_context.db,
            )

            async with graphql_context.db.start_session() as db:
                event = await cls._handle_decision(
                    db=db,
                    decision=data.decision,
                    proposed_change=proposed_change,
                    current_user=current_user,
                    context=graphql_context,
                )
                await proposed_change.save(db=db)

                if event:
                    event_service = await get_event_service()
                    await event_service.send(event=event)

        return {"ok": True}

    @classmethod
    async def _handle_decision(
        cls,
        db: InfrahubDatabase,
        decision: ProposedChangeApprovalDecision,
        proposed_change: CoreProposedChange,
        current_user: Node,
        context: GraphqlContext,
    ) -> InfrahubEvent | None:
        """Modify approved_by and rejected_by relationships of the prpoposed change based on the decision."""

        approved_by = await proposed_change.approved_by.get_peers(db=db)
        rejected_by = await proposed_change.rejected_by.get_peers(db=db)
        approved_by_ids = [node.id for _, node in approved_by.items()]
        rejected_by_ids = [node.id for _, node in rejected_by.items()]
        event: InfrahubEvent | None = None
        event_meta = EventMeta.from_context(context=context.get_context())

        match decision:
            case ProposedChangeApprovalDecision.APPROVE:
                if current_user.id in approved_by_ids:
                    raise ValidationError(input_value="You have already approved this proposed change")
                await proposed_change.approved_by.add(db=db, data=current_user)
                if current_user.id in rejected_by_ids:
                    await proposed_change.rejected_by.remove_locally(db=db, peer_id=current_user.id)

                event = ProposedChangeApprovedEvent(
                    proposed_change_id=proposed_change.id,
                    proposed_change_name=proposed_change.name.value,
                    proposed_change_state=proposed_change.state.value,
                    reviewer_account_id=current_user.id,
                    reviewer_account_name=current_user.name.value,
                    reviewer_decision=decision.value,
                    meta=event_meta,
                )

            case ProposedChangeApprovalDecision.CANCEL_APPROVE:
                if current_user.id not in approved_by_ids:
                    raise ValidationError(
                        input_value="You did not approve this proposed change yet, it can't be un-approved"
                    )
                await proposed_change.approved_by.remove_locally(db=db, peer_id=current_user.id)

                event = ProposedChangeApprovalRevokedEvent(
                    proposed_change_id=proposed_change.id,
                    proposed_change_name=proposed_change.name.value,
                    proposed_change_state=proposed_change.state.value,
                    reviewer_account_id=current_user.id,
                    reviewer_account_name=current_user.name.value,
                    reviewer_former_decision=ProposedChangeApprovalDecision.APPROVE.value,
                    meta=event_meta,
                )

            case ProposedChangeApprovalDecision.REJECT:
                if current_user.id in rejected_by_ids:
                    raise ValidationError(input_value="You have already rejected this proposed change")
                await proposed_change.rejected_by.add(db=db, data=current_user)
                if current_user.id in approved_by_ids:
                    await proposed_change.approved_by.remove_locally(db=db, peer_id=current_user.id)

                event = ProposedChangeRejectedEvent(
                    proposed_change_id=proposed_change.id,
                    proposed_change_name=proposed_change.name.value,
                    proposed_change_state=proposed_change.state.value,
                    reviewer_account_id=current_user.id,
                    reviewer_account_name=current_user.name.value,
                    reviewer_decision=decision.value,
                    meta=event_meta,
                )

            case ProposedChangeApprovalDecision.CANCEL_REJECT:
                if current_user.id not in rejected_by_ids:
                    raise ValidationError(
                        input_value="You did not reject this proposed change yet, it can't be un-rejected"
                    )
                await proposed_change.rejected_by.remove_locally(db=db, peer_id=current_user.id)

                event = ProposedChangeRejectionRevokedEvent(
                    proposed_change_id=proposed_change.id,
                    proposed_change_name=proposed_change.name.value,
                    proposed_change_state=proposed_change.state.value,
                    reviewer_account_id=current_user.id,
                    reviewer_account_name=current_user.name.value,
                    reviewer_former_decision=ProposedChangeApprovalDecision.REJECT.value,
                    meta=event_meta,
                )

            case _:
                raise ValidationError(input_value=f"Invalid decision {decision}")

        return event


class ProposedChangeMergeInput(InputObjectType):
    id = String(required=True)


class ProposedChangeMerge(Mutation):
    class Arguments:
        data = ProposedChangeMergeInput(required=True)
        wait_until_completion = Boolean(required=False)

    ok = Boolean()
    task = Field(TaskInfo, required=False)

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: dict[str, Any],
        wait_until_completion: bool = True,
    ) -> dict[str, bool]:
        graphql_context: GraphqlContext = info.context
        task: dict | None = None

        identifier = data.get("id", "")
        proposed_change = await NodeManager.get_one(
            id=identifier, kind=InfrahubKind.PROPOSEDCHANGE, db=graphql_context.db, raise_on_error=True
        )
        state = ProposedChangeState(proposed_change.state.value.value)
        if state != ProposedChangeState.OPEN:
            raise ValidationError("Only proposed change in OPEN state can be merged")

        async with graphql_context.db.start_session() as db:
            proposed_change.state.value = ProposedChangeState.MERGING.value
            await proposed_change.save(db=db)

        if wait_until_completion:
            await graphql_context.service.workflow.execute_workflow(
                workflow=PROPOSED_CHANGE_MERGE,
                context=graphql_context.get_context(),
                parameters={
                    "proposed_change_id": proposed_change.id,
                    "proposed_change_name": proposed_change.name.value,
                },
            )
        else:
            workflow = await graphql_context.service.workflow.submit_workflow(
                workflow=PROPOSED_CHANGE_MERGE,
                context=graphql_context.get_context(),
                parameters={
                    "proposed_change_id": proposed_change.id,
                    "proposed_change_name": proposed_change.name.value,
                },
            )
            task = {"id": workflow.id}

        return cls(ok=True, task=task)


class ProposedChangeCheckForApprovalRevokeInput(InputObjectType):
    ids = Field(List(of_type=String, required=True), required=False)


class ProposedChangeCheckForApprovalRevoke(Mutation):
    class Arguments:
        data = ProposedChangeCheckForApprovalRevokeInput(required=True)

    ok = Boolean()

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: dict[str, Any],
    ) -> dict[str, bool]:
        db = info.context.db
        ids: list[str] | None
        try:
            ids = data["ids"]
        except KeyError:
            ids = None

        await do_revoke_approvals_on_updated_pcs(db=db, proposed_changes_ids=ids)
        return cls(ok=True)


async def _get_source_branch(db: InfrahubDatabase, name: str) -> Branch:
    try:
        return await Branch.get_by_name(name=name, db=db)
    except BranchNotFoundError:
        raise ValidationError(
            input_value="The specified source branch for this proposed change was not found."
        ) from None



from __future__ import annotations

from typing import TYPE_CHECKING

from graphene import Field, ObjectType, String

from infrahub import __version__
from infrahub.core import registry

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo


class Info(ObjectType):
    deployment_id = String(required=True)
    version = String(required=True)

    @staticmethod
    async def resolve(
        root: dict,  # noqa: ARG004
        info: GraphQLResolveInfo,  # noqa: ARG004
    ) -> dict[str, str]:
        return {"deployment_id": str(registry.id), "version": __version__}


InfrahubInfo = Field(Info, resolver=Info.resolve, required=True)



from __future__ import annotations

import asyncio
import os
import sys
from enum import IntFlag
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from infrahub_sdk.exceptions import ModuleImportError
from infrahub_sdk.node import InfrahubNode
from infrahub_sdk.protocols import (
    CoreArtifactValidator,
    CoreGeneratorDefinition,
    CoreGeneratorValidator,
    CoreProposedChange,
)
from infrahub_sdk.schema.repository import InfrahubGeneratorDefinitionConfig
from prefect import flow, task
from prefect.cache_policies import NONE
from prefect.client.schemas.objects import (
    State,  # noqa: TC002
)
from prefect.logging import get_run_logger
from prefect.states import Completed, Failed
from pydantic import BaseModel

from infrahub import config, lock
from infrahub.artifacts.models import CheckArtifactCreate
from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.branch.tasks import merge_branch
from infrahub.core.constants import (
    CheckType,
    GeneratorInstanceStatus,
    InfrahubKind,
    RepositoryInternalStatus,
    ValidatorConclusion,
)
from infrahub.core.diff.coordinator import DiffCoordinator
from infrahub.core.diff.model.diff import DiffElementType, SchemaConflict
from infrahub.core.diff.model.path import NodeDiffFieldSummary
from infrahub.core.integrity.object_conflict.conflict_recorder import ObjectConflictValidatorRecorder
from infrahub.core.manager import NodeManager
from infrahub.core.protocols import CoreDataCheck, CoreValidator
from infrahub.core.protocols import CoreProposedChange as InternalCoreProposedChange
from infrahub.core.timestamp import Timestamp
from infrahub.core.validators.checks_runner import run_checks_and_update_validator
from infrahub.core.validators.determiner import ConstraintValidatorDeterminer
from infrahub.core.validators.models.validate_migration import SchemaValidateMigrationData
from infrahub.core.validators.tasks import schema_validate_migrations
from infrahub.dependencies.registry import get_component_registry
from infrahub.events import EventMeta, ProposedChangeMergedEvent
from infrahub.exceptions import MergeFailedError
from infrahub.generators.models import ProposedChangeGeneratorDefinition
from infrahub.git.base import extract_repo_file_information
from infrahub.git.models import TriggerRepositoryInternalChecks, TriggerRepositoryUserChecks
from infrahub.git.repository import InfrahubRepository, get_initialized_repo
from infrahub.log import get_logger
from infrahub.message_bus.types import (
    ProposedChangeArtifactDefinition,
    ProposedChangeBranchDiff,
    ProposedChangeRepository,
    ProposedChangeSubscriber,
)
from infrahub.proposed_change.branch_diff import (
    get_modified_node_ids,
    has_data_changes,
    has_node_changes,
    set_diff_summary_cache,
)
from infrahub.proposed_change.constants import ProposedChangeState
from infrahub.proposed_change.models import (
    RequestArtifactDefinitionCheck,
    RequestGeneratorDefinitionCheck,
    RequestProposedChangeDataIntegrity,
    RequestProposedChangePipeline,
    RequestProposedChangeRefreshArtifacts,
    RequestProposedChangeRepositoryChecks,
    RequestProposedChangeRunGenerators,
    RequestProposedChangeSchemaIntegrity,
    RequestProposedChangeUserTests,
    RunGeneratorAsCheckModel,
)
from infrahub.pytest_plugin import InfrahubBackendPlugin
from infrahub.validators.tasks import start_validator
from infrahub.workers.dependencies import get_cache, get_client, get_database, get_event_service, get_workflow
from infrahub.workflows.catalogue import (
    GIT_REPOSITORIES_CHECK_ARTIFACT_CREATE,
    GIT_REPOSITORY_INTERNAL_CHECKS_TRIGGER,
    GIT_REPOSITORY_USER_CHECKS_TRIGGER,
    REQUEST_ARTIFACT_DEFINITION_CHECK,
    REQUEST_GENERATOR_DEFINITION_CHECK,
    REQUEST_PROPOSED_CHANGE_DATA_INTEGRITY,
    REQUEST_PROPOSED_CHANGE_REFRESH_ARTIFACTS,
    REQUEST_PROPOSED_CHANGE_REPOSITORY_CHECKS,
    REQUEST_PROPOSED_CHANGE_RUN_GENERATORS,
    REQUEST_PROPOSED_CHANGE_SCHEMA_INTEGRITY,
    REQUEST_PROPOSED_CHANGE_USER_TESTS,
    RUN_GENERATOR_AS_CHECK,
)
from infrahub.workflows.utils import add_tags

from .branch_diff import get_diff_summary_cache, get_modified_kinds
from .checker import verify_proposed_change_is_mergeable

if TYPE_CHECKING:
    from infrahub_sdk.client import InfrahubClient
    from infrahub_sdk.diff import NodeDiff

    from infrahub.core.models import SchemaUpdateConstraintInfo
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.database import InfrahubDatabase


async def _proposed_change_transition_state(
    state: ProposedChangeState,
    database: InfrahubDatabase,
    proposed_change: InternalCoreProposedChange | None = None,
    proposed_change_id: str | None = None,
) -> None:
    async with database.start_session() as db:
        if proposed_change is None and proposed_change_id:
            proposed_change = await registry.manager.get_one(
                db=db, id=proposed_change_id, kind=InternalCoreProposedChange, raise_on_error=True
            )
        if proposed_change:
            proposed_change.state.value = state.value  # type: ignore[misc]
            await proposed_change.save(db=db)


# async def proposed_change_transition_merged(flow: Flow, flow_run: FlowRun, state: State) -> None:
#     await _proposed_change_transition_state(
#         proposed_change_id=flow_run.parameters["proposed_change_id"], state=ProposedChangeState.MERGED
#     )


# async def proposed_change_transition_open(flow: Flow, flow_run: FlowRun, state: State) -> None:
#     await _proposed_change_transition_state(
#         proposed_change_id=flow_run.parameters["proposed_change_id"], state=ProposedChangeState.OPEN
#     )


@flow(
    name="proposed-change-merge",
    flow_run_name="Merge propose change: {proposed_change_name} ",
    description="Merge a given proposed change.",
    # TODO need to investigate why these function are not working as expected
    # on_completion=[proposed_change_transition_merged],  # type: ignore
    # on_failure=[proposed_change_transition_open],  # type: ignore
    # on_crashed=[proposed_change_transition_open],  # type: ignore
    # on_cancellation=[proposed_change_transition_open],  # type: ignore
)
async def merge_proposed_change(
    proposed_change_id: str,
    proposed_change_name: str,  # noqa: ARG001
    context: InfrahubContext,
) -> State:
    log = get_run_logger()
    await add_tags(nodes=[proposed_change_id])
    database = await get_database()

    proposed_change = await registry.manager.get_one(
        db=database,
        id=proposed_change_id,
        kind=InternalCoreProposedChange,
        raise_on_error=True,
        prefetch_relationships=True,
    )

    async with database.start_session() as db:
        log.info("Validating if all conditions are met to merge the proposed change")

        try:
            await verify_proposed_change_is_mergeable(
                proposed_change=proposed_change,  # type: ignore[arg-type]
                db=db,
                account_session=context.account,
            )
        except ValueError as exc:
            await _proposed_change_transition_state(
                proposed_change=proposed_change, state=ProposedChangeState.OPEN, database=db
            )
            return Failed(message=str(exc))

        source_branch = await Branch.get_by_name(db=db, name=proposed_change.source_branch.value)
        validations = await proposed_change.validations.get_peers(db=db, peer_type=CoreValidator)
        for validation in validations.values():
            validator_kind = validation.get_kind()
            if (
                validator_kind != InfrahubKind.DATAVALIDATOR
                and validation.conclusion.value.value != ValidatorConclusion.SUCCESS.value
            ):
                # Ignoring Data integrity checks as they are handled again later
                await _proposed_change_transition_state(
                    proposed_change=proposed_change, state=ProposedChangeState.OPEN, database=db
                )
                return Failed(message="Unable to merge proposed change containing failing checks")
            if validator_kind == InfrahubKind.DATAVALIDATOR:
                data_checks = await validation.checks.get_peers(db=db, peer_type=CoreDataCheck)
                for check in data_checks.values():
                    if check.conflicts.value and not check.keep_branch.value:
                        await _proposed_change_transition_state(
                            proposed_change=proposed_change, state=ProposedChangeState.OPEN, database=db
                        )
                        return Failed(
                            message="Data conflicts found on branch and missing decisions about what branch to keep"
                        )

        log.info("Proposed change is eligible to be merged")
        try:
            await merge_branch(branch=source_branch.name, context=context, proposed_change_id=proposed_change_id)
        except MergeFailedError as exc:
            await _proposed_change_transition_state(
                proposed_change=proposed_change, state=ProposedChangeState.OPEN, database=db
            )
            return Failed(message=f"Merge failure when trying to merge {exc.message}")

        log.info(f"Branch {source_branch.name} has been merged successfully")

        await _proposed_change_transition_state(
            proposed_change=proposed_change, state=ProposedChangeState.MERGED, database=db
        )

        current_user = await NodeManager.get_one_by_id_or_default_filter(
            id=context.account.account_id, kind=InfrahubKind.GENERICACCOUNT, db=db
        )
        event_service = await get_event_service()
        await event_service.send(
            event=ProposedChangeMergedEvent(
                proposed_change_id=proposed_change.id,
                proposed_change_name=proposed_change.name.value,
                proposed_change_state=proposed_change.state.value,
                merged_by_account_id=current_user.id,
                merged_by_account_name=current_user.name.value,
                meta=EventMeta.from_context(context=context),
            )
        )

        return Completed(message="proposed change merged successfully")


@flow(
    name="proposed-changes-cancel-branch",
    flow_run_name="Cancel all proposed change associated with branch {branch_name}",
    description="Cancel all Proposed change associated with a branch.",
)
async def cancel_proposed_changes_branch(branch_name: str) -> None:
    await add_tags(branches=[branch_name])

    client = get_client()

    proposed_changed_opened = await client.filters(
        kind=CoreProposedChange,
        include=["id", "source_branch"],
        state__value=ProposedChangeState.OPEN.value,
        source_branch__value=branch_name,
    )
    proposed_changed_closed = await client.filters(
        kind=CoreProposedChange,
        include=["id", "source_branch"],
        state__value=ProposedChangeState.CLOSED.value,
        source_branch__value=branch_name,
    )

    for proposed_change in proposed_changed_opened + proposed_changed_closed:
        await cancel_proposed_change(proposed_change=proposed_change, client=get_client())


@task(name="Cancel a proposed change", description="Cancel a proposed change", cache_policy=NONE)  # type: ignore[arg-type]
async def cancel_proposed_change(proposed_change: CoreProposedChange, client: InfrahubClient) -> None:
    await add_tags(nodes=[proposed_change.id])
    log = get_run_logger()

    log.info("Canceling proposed change as the source branch was deleted")
    proposed_change = await client.get(kind=CoreProposedChange, id=proposed_change.id)
    proposed_change.state.value = ProposedChangeState.CANCELED.value
    await proposed_change.save()


@flow(name="proposed-changed-data-integrity", flow_run_name="Triggers data integrity check")
async def run_proposed_change_data_integrity_check(model: RequestProposedChangeDataIntegrity) -> None:
    """Triggers a data integrity validation check on the provided proposed change to start."""
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change])

    database = await get_database()
    async with database.start_session() as dbs:
        destination_branch = await registry.get_branch(db=dbs, branch=model.destination_branch)
        source_branch = await registry.get_branch(db=dbs, branch=model.source_branch)
        component_registry = get_component_registry()

        diff_coordinator = await component_registry.get_component(DiffCoordinator, db=dbs, branch=source_branch)
        await diff_coordinator.update_branch_diff(base_branch=destination_branch, diff_branch=source_branch)


@flow(name="proposed-changed-run-generator", flow_run_name="Run generators")
async def run_generators(model: RequestProposedChangeRunGenerators, context: InfrahubContext) -> None:
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change], db_change=True)

    client = get_client()

    generators = await client.filters(
        kind=CoreGeneratorDefinition,
        prefetch_relationships=True,
        populate_store=True,
        branch=model.source_branch,
    )
    generator_definitions = [
        ProposedChangeGeneratorDefinition(
            definition_id=generator.id,
            definition_name=generator.name.value,
            class_name=generator.class_name.value,
            file_path=generator.file_path.value,
            query_name=generator.query.peer.name.value,
            query_models=generator.query.peer.models.value,
            repository_id=generator.repository.peer.id,
            parameters=generator.parameters.value,
            group_id=generator.targets.peer.id,
            convert_query_response=generator.convert_query_response.value,
        )
        for generator in generators
    ]

    diff_summary = await get_diff_summary_cache(pipeline_id=model.branch_diff.pipeline_id)
    modified_kinds = get_modified_kinds(diff_summary=diff_summary, branch=model.source_branch)

    for generator_definition in generator_definitions:
        # Request generator definitions if the source branch that is managed in combination
        # to the Git repository containing modifications which could indicate changes to the transforms
        # in code
        # Alternatively if the queries used touches models that have been modified in the path
        # impacted artifact definitions will be included for consideration

        select = DefinitionSelect.NONE
        select = select.add_flag(
            current=select,
            flag=DefinitionSelect.FILE_CHANGES,
            condition=model.source_branch_sync_with_git and model.branch_diff.has_file_modifications,
        )

        for changed_model in modified_kinds:
            select = select.add_flag(
                current=select,
                flag=DefinitionSelect.MODIFIED_KINDS,
                condition=changed_model in generator_definition.query_models,
            )

        if select:
            request_generator_def_check_model = RequestGeneratorDefinitionCheck(
                generator_definition=generator_definition,
                branch_diff=model.branch_diff,
                proposed_change=model.proposed_change,
                source_branch=model.source_branch,
                source_branch_sync_with_git=model.source_branch_sync_with_git,
                destination_branch=model.destination_branch,
            )
            await get_workflow().submit_workflow(
                workflow=REQUEST_GENERATOR_DEFINITION_CHECK,
                parameters={"model": request_generator_def_check_model},
                context=context,
            )

    if model.refresh_artifacts:
        request_refresh_artifact_model = RequestProposedChangeRefreshArtifacts(
            proposed_change=model.proposed_change,
            source_branch=model.source_branch,
            source_branch_sync_with_git=model.source_branch_sync_with_git,
            destination_branch=model.destination_branch,
            branch_diff=model.branch_diff,
        )
        await get_workflow().submit_workflow(
            workflow=REQUEST_PROPOSED_CHANGE_REFRESH_ARTIFACTS,
            parameters={"model": request_refresh_artifact_model},
            context=context,
        )

    if model.do_repository_checks:
        model_proposed_change_repo_checks = RequestProposedChangeRepositoryChecks(
            proposed_change=model.proposed_change,
            source_branch=model.source_branch,
            source_branch_sync_with_git=model.source_branch_sync_with_git,
            destination_branch=model.destination_branch,
            branch_diff=model.branch_diff,
        )
        await get_workflow().submit_workflow(
            workflow=REQUEST_PROPOSED_CHANGE_REPOSITORY_CHECKS,
            context=context,
            parameters={"model": model_proposed_change_repo_checks},
        )


@flow(name="proposed-changed-schema-integrity", flow_run_name="Process schema integrity")
async def run_proposed_change_schema_integrity_check(model: RequestProposedChangeSchemaIntegrity) -> None:
    # For now, we retrieve the latest schema for each branch from the registry
    # In the future it would be good to generate the object SchemaUpdateValidationResult from message.branch_diff
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change])

    source_schema = registry.schema.get_schema_branch(name=model.source_branch).duplicate()
    dest_schema = registry.schema.get_schema_branch(name=model.destination_branch).duplicate()

    candidate_schema = dest_schema.duplicate()
    candidate_schema.update(schema=source_schema)
    schema_diff = dest_schema.diff(other=candidate_schema)
    validation_result = dest_schema.validate_update(other=candidate_schema, diff=schema_diff)

    diff_summary = await get_diff_summary_cache(pipeline_id=model.branch_diff.pipeline_id)
    constraints_from_data_diff = await _get_proposed_change_schema_integrity_constraints(
        schema=candidate_schema, diff_summary=diff_summary
    )
    constraints_from_schema_diff = validation_result.constraints
    constraints = set(constraints_from_data_diff + constraints_from_schema_diff)

    if not constraints:
        return

    # ----------------------------------------------------------
    # Validate if the new schema is valid with the content of the database
    # ----------------------------------------------------------
    source_branch = registry.get_branch_from_registry(branch=model.source_branch)
    responses = await schema_validate_migrations(
        message=SchemaValidateMigrationData(
            branch=source_branch, schema_branch=candidate_schema, constraints=list(constraints)
        )
    )

    # TODO we need to report a failure if an error happened during the execution of a validator
    conflicts: list[SchemaConflict] = []
    for response in responses:
        for violation in response.violations:
            conflicts.append(
                SchemaConflict(
                    name=response.schema_path.get_path(),
                    type=response.constraint_name,
                    kind=violation.node_kind,
                    id=violation.node_id,
                    path=response.schema_path.get_path(),
                    value=violation.message,
                    branch="placeholder",
                )
            )

    if not conflicts:
        return

    database = await get_database()
    async with database.start_transaction() as db:
        object_conflict_validator_recorder = ObjectConflictValidatorRecorder(
            db=db,
            validator_kind=InfrahubKind.SCHEMAVALIDATOR,
            validator_label="Schema Integrity",
            check_schema_kind=InfrahubKind.SCHEMACHECK,
        )
        await object_conflict_validator_recorder.record_conflicts(
            proposed_change_id=model.proposed_change, conflicts=conflicts
        )


async def _get_proposed_change_schema_integrity_constraints(
    schema: SchemaBranch, diff_summary: list[NodeDiff]
) -> list[SchemaUpdateConstraintInfo]:
    node_diff_field_summary_map: dict[str, NodeDiffFieldSummary] = {}

    for node_diff in diff_summary:
        node_kind = node_diff["kind"]
        if node_kind not in node_diff_field_summary_map:
            node_diff_field_summary_map[node_kind] = NodeDiffFieldSummary(kind=node_kind)
        field_summary = node_diff_field_summary_map[node_kind]
        for element in node_diff["elements"]:
            element_name = element["name"]
            element_type = element["element_type"]
            if element_type.lower() in (
                DiffElementType.RELATIONSHIP_MANY.value.lower(),
                DiffElementType.RELATIONSHIP_ONE.value.lower(),
            ):
                field_summary.relationship_names.add(element_name)
            elif element_type.lower() in (DiffElementType.ATTRIBUTE.value.lower(),):
                field_summary.attribute_names.add(element_name)

    determiner = ConstraintValidatorDeterminer(schema_branch=schema)
    return await determiner.get_constraints(node_diffs=list(node_diff_field_summary_map.values()))


@flow(name="proposed-changed-repository-checks", flow_run_name="Process user defined checks")
async def repository_checks(model: RequestProposedChangeRepositoryChecks, context: InfrahubContext) -> None:
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change])

    for repository in model.branch_diff.repositories:
        if (
            model.source_branch_sync_with_git
            and not repository.read_only
            and repository.internal_status == RepositoryInternalStatus.ACTIVE.value
        ):
            trigger_internal_checks_model = TriggerRepositoryInternalChecks(
                proposed_change=model.proposed_change,
                repository=repository.repository_id,
                source_branch=model.source_branch,
                target_branch=model.destination_branch,
            )
            await get_workflow().submit_workflow(
                workflow=GIT_REPOSITORY_INTERNAL_CHECKS_TRIGGER,
                context=context,
                parameters={"model": trigger_internal_checks_model},
            )

        trigger_user_checks_model = TriggerRepositoryUserChecks(
            proposed_change=model.proposed_change,
            repository_id=repository.repository_id,
            repository_name=repository.repository_name,
            source_branch=model.source_branch,
            source_branch_sync_with_git=model.source_branch_sync_with_git,
            target_branch=model.destination_branch,
            branch_diff=model.branch_diff,
        )
        await get_workflow().submit_workflow(
            workflow=GIT_REPOSITORY_USER_CHECKS_TRIGGER,
            context=context,
            parameters={"model": trigger_user_checks_model},
        )


@flow(name="proposed-changed-user-tests", flow_run_name="Run unit tests in repositories")
async def run_proposed_change_user_tests(model: RequestProposedChangeUserTests) -> None:
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change])

    log = get_run_logger()
    client = get_client()

    proposed_change = await client.get(kind=InfrahubKind.PROPOSEDCHANGE, id=model.proposed_change)

    def _execute(
        directory: Path, repository: ProposedChangeRepository, proposed_change: InfrahubNode
    ) -> int | pytest.ExitCode:
        # Check for both .infrahub.yml and .infrahub.yaml, prefer .yml if both exist
        config_file_yml = directory / ".infrahub.yml"
        config_file_yaml = directory / ".infrahub.yaml"

        if config_file_yml.is_file():
            config_file = str(config_file_yml)
        elif config_file_yaml.is_file():
            config_file = str(config_file_yaml)
        else:
            config_file = str(config_file_yml)  # Default to .yml for error messages
        test_directory = directory / "tests"
        log = get_logger()

        if not test_directory.is_dir():
            log.debug(
                event="repository_tests_ignored",
                proposed_change=proposed_change,
                repository=repository.repository_name,
                message="tests directory not found",
            )
            return 1

        # Check if config file exists and log error if neither extension is found
        if not config_file_yml.is_file() and not config_file_yaml.is_file():
            log.error(
                event="repository_tests_failed",
                proposed_change=proposed_change,
                repository=repository.repository_name,
                message="Configuration file not found (.infrahub.yml or .infrahub.yaml)",
            )
            return 1

        # Redirect stdout/stderr to avoid showing pytest lines in the git agent
        old_out = sys.stdout
        old_err = sys.stderr

        with Path(os.devnull).open(mode="w", encoding="utf-8") as devnull:
            sys.stdout = devnull
            sys.stderr = devnull

            exit_code = pytest.main(
                [
                    str(test_directory),
                    f"--infrahub-repo-config={config_file}",
                    f"--infrahub-address={config.SETTINGS.main.internal_address}",
                    "-qqqq",
                    "-s",
                ],
                plugins=[InfrahubBackendPlugin(client.config, repository.repository_id, proposed_change.id)],
            )

        # Restore stdout/stderr back to their orignal states
        sys.stdout = old_out
        sys.stderr = old_err

        return exit_code

    for repository in model.branch_diff.repositories:
        if model.source_branch_sync_with_git:
            repo = await get_initialized_repo(
                client=client,
                repository_id=repository.repository_id,
                name=repository.repository_name,
                repository_kind=repository.kind,
            )
            commit = repo.get_commit_value(proposed_change.source_branch.value)
            worktree_directory = Path(repo.get_commit_worktree(commit=commit).directory)

            return_code = await asyncio.to_thread(_execute, worktree_directory, repository, proposed_change)
            log.info(msg=f"repository_tests_completed return_code={return_code}")


@flow(
    name="artifacts-generation-validation",
    flow_run_name="Validating generation of artifacts for {model.artifact_definition.definition_name}",
)
async def validate_artifacts_generation(model: RequestArtifactDefinitionCheck, context: InfrahubContext) -> None:
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change], db_change=True)

    log = get_run_logger()
    client = get_client()

    artifact_definition = await client.get(
        kind=InfrahubKind.ARTIFACTDEFINITION,
        id=model.artifact_definition.definition_id,
        branch=model.source_branch,
    )
    proposed_change = await client.get(kind=InfrahubKind.PROPOSEDCHANGE, id=model.proposed_change)

    validator_name = f"Artifact Validator: {model.artifact_definition.definition_name}"

    await proposed_change.validations.fetch()

    previous_validator: CoreArtifactValidator | None = None
    for relationship in proposed_change.validations.peers:
        existing_validator = relationship.peer
        if (
            existing_validator.typename == InfrahubKind.ARTIFACTVALIDATOR
            and existing_validator.definition.id == model.artifact_definition.definition_id
        ):
            previous_validator = existing_validator

    validator = await start_validator(
        client=client,
        validator=previous_validator,
        validator_type=CoreArtifactValidator,
        proposed_change=model.proposed_change,
        data={
            "label": validator_name,
            "definition": model.artifact_definition.definition_id,
        },
        context=context,
    )

    await artifact_definition.targets.fetch()
    group = artifact_definition.targets.peer
    await group.members.fetch()

    existing_artifacts = await client.filters(
        kind=InfrahubKind.ARTIFACT,
        definition__ids=[model.artifact_definition.definition_id],
        include=["object"],
        branch=model.source_branch,
    )
    artifacts_by_member = {}
    for artifact in existing_artifacts:
        artifacts_by_member[artifact.object.peer.id] = artifact.id

    repository = model.branch_diff.get_repository(repository_id=model.artifact_definition.repository_id)
    impacted_artifacts = model.branch_diff.get_subscribers_ids(kind=InfrahubKind.ARTIFACT)

    checks = []

    for relationship in group.members.peers:
        member = relationship.peer
        artifact_id = artifacts_by_member.get(member.id)
        if _should_render_artifact(
            artifact_id=artifact_id,
            managed_branch=model.source_branch_sync_with_git,
            impacted_artifacts=impacted_artifacts,
        ):
            log.info(f"Trigger Artifact processing for {member.display_label}")

            check_model = CheckArtifactCreate(
                context=context,
                artifact_name=model.artifact_definition.artifact_name,
                artifact_id=artifact_id,
                artifact_definition=model.artifact_definition.definition_id,
                commit=repository.source_commit,
                content_type=model.artifact_definition.content_type,
                transform_type=model.artifact_definition.transform_kind,
                transform_location=model.artifact_definition.transform_location,
                convert_query_response=model.artifact_definition.convert_query_response,
                repository_id=repository.repository_id,
                repository_name=repository.repository_name,
                repository_kind=repository.kind,
                branch_name=model.source_branch,
                query=model.artifact_definition.query_name,
                variables=await member.extract(params=artifact_definition.parameters.value),
                target_id=member.id,
                target_kind=member.get_kind(),
                target_name=member.display_label,
                timeout=model.artifact_definition.timeout,
                validator_id=validator.id,
            )

            checks.append(
                get_workflow().execute_workflow(
                    workflow=GIT_REPOSITORIES_CHECK_ARTIFACT_CREATE,
                    parameters={"model": check_model},
                    expected_return=ValidatorConclusion,
                )
            )

    await run_checks_and_update_validator(
        event_service=await get_event_service(),
        checks=checks,
        validator=validator,
        proposed_change_id=model.proposed_change,
        context=context,
    )


def _should_render_artifact(artifact_id: str | None, managed_branch: bool, impacted_artifacts: list[str]) -> bool:  # noqa: ARG001
    """Returns a boolean to indicate if an artifact should be generated or not.
    Will return true if:
        * The artifact_id wasn't set which could be that it's a new object that doesn't have a previous artifact
        * The source brance is not data only which would indicate that it could contain updates in git to the transform
        * The artifact_id exists in the impacted_artifacts list
    Will return false if:
        * The source branch is a data only branch and the artifact_id exists and is not in the impacted list
    """

    # if not artifact_id or managed_branch:
    #    return True
    # return artifact_id in impacted_artifacts
    # Temporary workaround tracked in https://github.com/opsmill/infrahub/issues/4991
    return True


@flow(
    name="run-generator-as-check",
    flow_run_name="Execute Generator {model.generator_definition.definition_name} for {model.target_name}",
)
async def run_generator_as_check(model: RunGeneratorAsCheckModel, context: InfrahubContext) -> ValidatorConclusion:
    await add_tags(branches=[model.branch_name], nodes=[model.proposed_change], db_change=True)

    client = get_client()
    log = get_run_logger()

    repository = await get_initialized_repo(
        client=client,
        repository_id=model.repository_id,
        name=model.repository_name,
        repository_kind=model.repository_kind,
        commit=model.commit,
    )

    conclusion = ValidatorConclusion.SUCCESS

    generator_definition = InfrahubGeneratorDefinitionConfig(
        name=model.generator_definition.definition_name,
        class_name=model.generator_definition.class_name,
        file_path=model.generator_definition.file_path,
        query=model.generator_definition.query_name,
        targets=model.generator_definition.group_id,
        convert_query_response=model.generator_definition.convert_query_response,
    )

    commit_worktree = repository.get_commit_worktree(commit=model.commit)

    file_info = extract_repo_file_information(
        full_filename=commit_worktree.directory / generator_definition.file_path,
        repo_directory=repository.directory_root,
        worktree_directory=commit_worktree.directory,
    )
    generator_instance = await _define_instance(model=model, client=client)

    check_message = "Instance successfully generated"
    try:
        log.debug(f"repo information {file_info}")
        log.debug(f"Root directory : {repository.directory_root}")
        generator_class = generator_definition.load_class(
            import_root=repository.directory_root, relative_path=file_info.relative_repo_path_dir
        )

        generator = generator_class(
            query=generator_definition.query,
            client=client,
            branch=model.branch_name,
            params=model.variables,
            generator_instance=generator_instance.id,
            convert_query_response=generator_definition.convert_query_response,
            infrahub_node=InfrahubNode,
        )
        generator._init_client.request_context = context.to_request_context()
        await generator.run(identifier=generator_definition.name)
        generator_instance.status.value = GeneratorInstanceStatus.READY.value
    except ModuleImportError as exc:
        conclusion = ValidatorConclusion.FAILURE
        generator_instance.status.value = GeneratorInstanceStatus.ERROR.value
        check_message = f"Failed to import generator: {exc.message}"
        log.exception(check_message, exc_info=exc)
    except Exception as exc:
        conclusion = ValidatorConclusion.FAILURE
        generator_instance.status.value = GeneratorInstanceStatus.ERROR.value
        check_message = f"Failed to execute generator: {str(exc)}"
        log.exception(check_message, exc_info=exc)

    log.info("Generator run completed, starting update")
    await generator_instance.update(do_full_update=True)

    check = None
    existing_check = await client.filters(
        kind=InfrahubKind.GENERATORCHECK, validator__ids=model.validator_id, instance__value=generator_instance.id
    )
    if existing_check:
        check = existing_check[0]

    if check:
        check.created_at.value = Timestamp().to_string()
        check.conclusion.value = conclusion.value
        await check.save()
    else:
        check = await client.create(
            kind=InfrahubKind.GENERATORCHECK,
            data={
                "name": model.target_name,
                "origin": model.repository_id,
                "kind": "GeneratorDefinition",
                "validator": model.validator_id,
                "created_at": Timestamp().to_string(),
                "message": check_message,
                "conclusion": conclusion.value,
                "instance": generator_instance.id,
            },
        )
        await check.save()

    return conclusion


async def _define_instance(model: RunGeneratorAsCheckModel, client: InfrahubClient) -> InfrahubNode:
    if model.generator_instance:
        instance = await client.get(
            kind=InfrahubKind.GENERATORINSTANCE, id=model.generator_instance, branch=model.branch_name
        )
        instance.status.value = GeneratorInstanceStatus.PENDING.value
        await instance.update(do_full_update=True)

    else:
        async with lock.registry.get(
            f"{model.target_id}-{model.generator_definition.definition_id}", namespace="generator"
        ):
            instances = await client.filters(
                kind=InfrahubKind.GENERATORINSTANCE,
                definition__ids=[model.generator_definition.definition_id],
                object__ids=[model.target_id],
                branch=model.branch_name,
            )
            if instances:
                instance = instances[0]
                instance.status.value = GeneratorInstanceStatus.PENDING.value
                await instance.update(do_full_update=True)
            else:
                instance = await client.create(
                    kind=InfrahubKind.GENERATORINSTANCE,
                    branch=model.branch_name,
                    data={
                        "name": f"{model.generator_definition.definition_name}: {model.target_name}",
                        "status": GeneratorInstanceStatus.PENDING.value,
                        "object": model.target_id,
                        "definition": model.generator_definition.definition_id,
                    },
                )
                await instance.save()
    return instance


@flow(
    name="request-generator-definition-check",
    flow_run_name="Validate Generator selection for {model.generator_definition.definition_name}",
)
async def request_generator_definition_check(model: RequestGeneratorDefinitionCheck, context: InfrahubContext) -> None:
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change])

    log = get_run_logger()
    client = get_client()

    proposed_change = await client.get(kind=InfrahubKind.PROPOSEDCHANGE, id=model.proposed_change)

    validator_name = f"Generator Validator: {model.generator_definition.definition_name}"
    await proposed_change.validations.fetch()

    previous_validator: CoreGeneratorValidator | None = None
    for relationship in proposed_change.validations.peers:
        existing_validator = relationship.peer
        if (
            existing_validator.typename == InfrahubKind.GENERATORVALIDATOR
            and existing_validator.definition.id == model.generator_definition.definition_id
        ):
            previous_validator = existing_validator

    validator = await start_validator(
        client=client,
        validator=previous_validator,
        validator_type=CoreGeneratorValidator,
        proposed_change=model.proposed_change,
        data={
            "label": validator_name,
            "definition": model.generator_definition.definition_id,
        },
        context=context,
    )

    group = await client.get(
        kind=InfrahubKind.GENERICGROUP,
        prefetch_relationships=True,
        populate_store=True,
        id=model.generator_definition.group_id,
        branch=model.source_branch,
    )
    await group.members.fetch()

    existing_instances = await client.filters(
        kind=InfrahubKind.GENERATORINSTANCE,
        definition__ids=[model.generator_definition.definition_id],
        include=["object"],
        branch=model.source_branch,
    )
    instance_by_member = {}
    for instance in existing_instances:
        instance_by_member[instance.object.peer.id] = instance.id

    repository = model.branch_diff.get_repository(repository_id=model.generator_definition.repository_id)
    requested_instances = 0
    impacted_instances = model.branch_diff.get_subscribers_ids(kind=InfrahubKind.GENERATORINSTANCE)

    check_generator_run_models = []
    for relationship in group.members.peers:
        member = relationship.peer
        generator_instance = instance_by_member.get(member.id)
        if _run_generator(
            instance_id=generator_instance,
            managed_branch=model.source_branch_sync_with_git,
            impacted_instances=impacted_instances,
        ):
            requested_instances += 1
            log.info(f"Trigger execution of {model.generator_definition.definition_name} for {member.display_label}")
            check_generator_run_model = RunGeneratorAsCheckModel(
                generator_definition=model.generator_definition,
                generator_instance=generator_instance,
                commit=repository.source_commit,
                repository_id=repository.repository_id,
                repository_name=repository.repository_name,
                repository_kind=repository.kind,
                branch_name=model.source_branch,
                query=model.generator_definition.query_name,
                variables=await member.extract(params=model.generator_definition.parameters),
                target_id=member.id,
                target_name=member.display_label,
                validator_id=validator.id,
                proposed_change=model.proposed_change,
            )
            check_generator_run_models.append(check_generator_run_model)

    checks_coroutines = [
        get_workflow().execute_workflow(
            workflow=RUN_GENERATOR_AS_CHECK,
            parameters={"model": check_generator_run_model},
            expected_return=ValidatorConclusion,
            context=context,
        )
        for check_generator_run_model in check_generator_run_models
    ]

    await run_checks_and_update_validator(
        event_service=await get_event_service(),
        checks=checks_coroutines,
        validator=validator,
        context=context,
        proposed_change_id=proposed_change.id,
    )


def _run_generator(instance_id: str | None, managed_branch: bool, impacted_instances: list[str]) -> bool:
    """Returns a boolean to indicate if a generator instance needs to be executed
    Will return true if:
        * The instance_id wasn't set which could be that it's a new object that doesn't have a previous generator instance
        * The source branch is set to sync with Git which would indicate that it could contain updates in git to the generator
        * The instance_id exists in the impacted_instances list
    Will return false if:
        * The source branch is a not one that syncs with git and the instance_id exists and is not in the impacted list
    """
    if not instance_id or managed_branch:
        return True
    return instance_id in impacted_instances


class DefinitionSelect(IntFlag):
    NONE = 0
    MODIFIED_KINDS = 1
    FILE_CHANGES = 2

    @staticmethod
    def add_flag(current: DefinitionSelect, flag: DefinitionSelect, condition: bool) -> DefinitionSelect:
        if condition:
            return current | flag
        return current

    @property
    def log_line(self) -> str:
        change_types = []
        if DefinitionSelect.MODIFIED_KINDS in self:
            change_types.append("data changes within relevant object kinds")

        if DefinitionSelect.FILE_CHANGES in self:
            change_types.append("file modifications in Git repositories")

        if self:
            return f"Requesting generation due to {' and '.join(change_types)}"

        return "Doesn't require changes due to no relevant modified kinds or file changes in Git"


@flow(name="proposed-changed-pipeline", flow_run_name="Execute proposed changed pipeline")
async def run_proposed_change_pipeline(model: RequestProposedChangePipeline, context: InfrahubContext) -> None:
    client = get_client()
    repositories = await _get_proposed_change_repositories(model=model, client=client)

    if model.source_branch_sync_with_git and await _validate_repository_merge_conflicts(
        repositories=repositories, client=client
    ):
        for repo in repositories:
            if not repo.read_only and repo.internal_status == RepositoryInternalStatus.ACTIVE.value:
                trigger_repo_checks_model = TriggerRepositoryInternalChecks(
                    proposed_change=model.proposed_change,
                    repository=repo.repository_id,
                    source_branch=repo.source_branch,
                    target_branch=repo.destination_branch,
                )
                await get_workflow().submit_workflow(
                    workflow=GIT_REPOSITORY_INTERNAL_CHECKS_TRIGGER,
                    context=context,
                    parameters={"model": trigger_repo_checks_model},
                )
        return

    await _gather_repository_repository_diffs(repositories=repositories, client=client)

    database = await get_database()
    async with database.start_session() as dbs:
        destination_branch = await registry.get_branch(db=dbs, branch=model.destination_branch)
        source_branch = await registry.get_branch(db=dbs, branch=model.source_branch)
        component_registry = get_component_registry()
        diff_coordinator = await component_registry.get_component(DiffCoordinator, db=dbs, branch=source_branch)
        await diff_coordinator.update_branch_diff(base_branch=destination_branch, diff_branch=source_branch)

    client = get_client()

    diff_summary = await client.get_diff_summary(branch=model.source_branch)
    await set_diff_summary_cache(pipeline_id=model.pipeline_id, diff_summary=diff_summary, cache=await get_cache())
    branch_diff = ProposedChangeBranchDiff(pipeline_id=model.pipeline_id, repositories=repositories)
    await _populate_subscribers(
        branch_diff=branch_diff, diff_summary=diff_summary, branch=model.source_branch, client=client
    )

    if model.check_type is CheckType.ARTIFACT:
        request_refresh_artifact_model = RequestProposedChangeRefreshArtifacts(
            proposed_change=model.proposed_change,
            source_branch=model.source_branch,
            source_branch_sync_with_git=model.source_branch_sync_with_git,
            destination_branch=model.destination_branch,
            branch_diff=branch_diff,
        )
        await get_workflow().submit_workflow(
            workflow=REQUEST_PROPOSED_CHANGE_REFRESH_ARTIFACTS,
            parameters={"model": request_refresh_artifact_model},
            context=context,
        )

    if model.check_type in [CheckType.ALL, CheckType.GENERATOR]:
        model_proposed_change_run_generator = RequestProposedChangeRunGenerators(
            proposed_change=model.proposed_change,
            source_branch=model.source_branch,
            source_branch_sync_with_git=model.source_branch_sync_with_git,
            destination_branch=model.destination_branch,
            branch_diff=branch_diff,
            refresh_artifacts=model.check_type is CheckType.ALL,
            do_repository_checks=model.check_type is CheckType.ALL,
        )
        await get_workflow().submit_workflow(
            workflow=REQUEST_PROPOSED_CHANGE_RUN_GENERATORS,
            context=context,
            parameters={"model": model_proposed_change_run_generator},
        )

    if model.check_type in [CheckType.ALL, CheckType.DATA] and has_node_changes(
        diff_summary=diff_summary, branch=model.source_branch
    ):
        model_proposed_change_data_integrity = RequestProposedChangeDataIntegrity(
            proposed_change=model.proposed_change,
            source_branch=model.source_branch,
            source_branch_sync_with_git=model.source_branch_sync_with_git,
            destination_branch=model.destination_branch,
            branch_diff=branch_diff,
        )
        await get_workflow().submit_workflow(
            workflow=REQUEST_PROPOSED_CHANGE_DATA_INTEGRITY,
            context=context,
            parameters={"model": model_proposed_change_data_integrity},
        )

    if model.check_type in [CheckType.REPOSITORY, CheckType.USER]:
        model_proposed_change_repo_checks = RequestProposedChangeRepositoryChecks(
            proposed_change=model.proposed_change,
            source_branch=model.source_branch,
            source_branch_sync_with_git=model.source_branch_sync_with_git,
            destination_branch=model.destination_branch,
            branch_diff=branch_diff,
        )
        await get_workflow().submit_workflow(
            workflow=REQUEST_PROPOSED_CHANGE_REPOSITORY_CHECKS,
            context=context,
            parameters={"model": model_proposed_change_repo_checks},
        )

    if model.check_type in [CheckType.ALL, CheckType.SCHEMA] and has_data_changes(
        diff_summary=diff_summary, branch=model.source_branch
    ):
        await get_workflow().submit_workflow(
            workflow=REQUEST_PROPOSED_CHANGE_SCHEMA_INTEGRITY,
            context=context,
            parameters={
                "model": RequestProposedChangeSchemaIntegrity(
                    proposed_change=model.proposed_change,
                    source_branch=model.source_branch,
                    source_branch_sync_with_git=model.source_branch_sync_with_git,
                    destination_branch=model.destination_branch,
                    branch_diff=branch_diff,
                )
            },
        )

    if model.check_type in [CheckType.ALL, CheckType.TEST]:
        await get_workflow().submit_workflow(
            workflow=REQUEST_PROPOSED_CHANGE_USER_TESTS,
            context=context,
            parameters={
                "model": RequestProposedChangeUserTests(
                    proposed_change=model.proposed_change,
                    source_branch=model.source_branch,
                    source_branch_sync_with_git=model.source_branch_sync_with_git,
                    destination_branch=model.destination_branch,
                    branch_diff=branch_diff,
                )
            },
        )


@flow(
    name="proposed-changed-refresh-artifacts",
    flow_run_name="Trigger artifacts refresh",
)
async def refresh_artifacts(model: RequestProposedChangeRefreshArtifacts, context: InfrahubContext) -> None:
    await add_tags(branches=[model.source_branch], nodes=[model.proposed_change])
    log = get_run_logger()

    client = get_client()

    definition_information = await client.execute_graphql(
        query=GATHER_ARTIFACT_DEFINITIONS,
        branch_name=model.source_branch,
    )
    artifact_definitions = _parse_artifact_definitions(
        definitions=definition_information[InfrahubKind.ARTIFACTDEFINITION]["edges"]
    )
    diff_summary = await get_diff_summary_cache(pipeline_id=model.branch_diff.pipeline_id)
    modified_kinds = get_modified_kinds(diff_summary=diff_summary, branch=model.source_branch)

    for artifact_definition in artifact_definitions:
        # Request artifact definition checks if the source branch that is managed in combination
        # to the Git repository containing modifications which could indicate changes to the transforms
        # in code
        # Alternatively if the queries used touches models that have been modified in the path
        # impacted artifact definitions will be included for consideration

        select = DefinitionSelect.NONE
        select = select.add_flag(
            current=select,
            flag=DefinitionSelect.FILE_CHANGES,
            condition=model.source_branch_sync_with_git and model.branch_diff.has_file_modifications,
        )

        for changed_model in modified_kinds:
            condition = False
            if (changed_model in artifact_definition.query_models) or (
                changed_model.startswith("Profile")
                and changed_model.replace("Profile", "", 1) in artifact_definition.query_models
            ):
                condition = True

            select = select.add_flag(
                current=select,
                flag=DefinitionSelect.MODIFIED_KINDS,
                condition=condition,
            )

        if select:
            log.info(f"Trigger processing of {artifact_definition.definition_name}")
            request_artifacts_definitions_model = RequestArtifactDefinitionCheck(
                artifact_definition=artifact_definition,
                branch_diff=model.branch_diff,
                proposed_change=model.proposed_change,
                source_branch=model.source_branch,
                source_branch_sync_with_git=model.source_branch_sync_with_git,
                destination_branch=model.destination_branch,
            )

            await get_workflow().submit_workflow(
                REQUEST_ARTIFACT_DEFINITION_CHECK,
                parameters={"model": request_artifacts_definitions_model},
                context=context,
            )


GATHER_ARTIFACT_DEFINITIONS = """
query GatherArtifactDefinitions {
  CoreArtifactDefinition {
    edges {
      node {
        id
        name {
          value
        }
        artifact_name {
          value
        }
        content_type {
            value
        }
        transformation {
          node {
            __typename
            timeout {
                value
            }
            query {
              node {
                models {
                  value
                }
                name {
                  value
                }
              }
            }
            ... on CoreTransformJinja2 {
              template_path {
                value
              }
            }
            ... on CoreTransformPython {
              class_name {
                value
              }
              file_path {
                value
              }
              convert_query_response {
                value
              }
            }
            repository {
              node {
                id
              }
            }
          }
        }
      }
    }
  }
}
"""

GATHER_GRAPHQL_QUERY_SUBSCRIBERS = """
query GatherGraphQLQuerySubscribers($members: [ID!]) {
  CoreGraphQLQueryGroup(members__ids: $members) {
    edges {
      node {
        subscribers {
          edges {
            node {
              id
              __typename
            }
          }
        }
      }
    }
  }
}
"""


DESTINATION_ALLREPOSITORIES = """
query DestinationBranchRepositories {
  CoreGenericRepository {
    edges {
      node {
        __typename
        id
        name {
          value
        }
        internal_status {
          value
        }
        ... on CoreRepository {
          commit {
            value
          }
        }
        ... on CoreReadOnlyRepository {
          commit {
            value
          }
        }
      }
    }
  }
}
"""

SOURCE_REPOSITORIES = """
query MyQuery {
  CoreRepository {
    edges {
      node {
        __typename
        id
        name {
          value
        }
        internal_status {
          value
        }
        commit {
          value
        }
      }
    }
  }
}
"""
SOURCE_READONLY_REPOSITORIES = """
query MyQuery {
  CoreReadOnlyRepository {
    edges {
      node {
        __typename
        id
        name {
          value
        }
        internal_status {
          value
        }
        commit {
          value
        }
      }
    }
  }
}
"""


class Repository(BaseModel):
    repository_id: str
    repository_name: str
    read_only: bool
    commit: str
    internal_status: str


def _parse_proposed_change_repositories(
    model: RequestProposedChangePipeline, source: list[dict], destination: list[dict]
) -> list[ProposedChangeRepository]:
    """This function assumes that the repos is a list of the edges

    The data should come from the queries:
    * DESTINATION_ALLREPOSITORIES
    * SOURCE_REPOSITORIES
    * SOURCE_READONLY_REPOSITORIES
    """
    destination_repos = _parse_repositories(repositories=destination)
    source_repos = _parse_repositories(repositories=source)
    pc_repos: dict[str, ProposedChangeRepository] = {}
    for repo in destination_repos:
        if repo.repository_id not in pc_repos:
            pc_repos[repo.repository_id] = ProposedChangeRepository(
                repository_id=repo.repository_id,
                repository_name=repo.repository_name,
                read_only=repo.read_only,
                internal_status=repo.internal_status,
                destination_commit=repo.commit,
                source_branch=model.source_branch,
                destination_branch=model.destination_branch,
            )
        else:
            pc_repos[repo.repository_id].destination_commit = repo.commit

    for repo in source_repos:
        if repo.repository_id not in pc_repos:
            pc_repos[repo.repository_id] = ProposedChangeRepository(
                repository_id=repo.repository_id,
                repository_name=repo.repository_name,
                read_only=repo.read_only,
                internal_status=repo.internal_status,
                source_commit=repo.commit,
                source_branch=model.source_branch,
                destination_branch=model.destination_branch,
            )
        else:
            pc_repos[repo.repository_id].source_commit = repo.commit
            pc_repos[repo.repository_id].internal_status = repo.internal_status

    return list(pc_repos.values())


def _parse_repositories(repositories: list[dict]) -> list[Repository]:
    """This function assumes that the repos is a list of the edges

    The data should come from the queries:
    * DESTINATION_ALLREPOSITORIES
    * SOURCE_REPOSITORIES
    * SOURCE_READONLY_REPOSITORIES
    """
    parsed = []
    for repo in repositories:
        parsed.append(
            Repository(
                repository_id=repo["node"]["id"],
                repository_name=repo["node"]["name"]["value"],
                read_only=repo["node"]["__typename"] == InfrahubKind.READONLYREPOSITORY,
                commit=repo["node"]["commit"]["value"] or "",
                internal_status=repo["node"]["internal_status"]["value"],
            )
        )
    return parsed


def _parse_artifact_definitions(definitions: list[dict]) -> list[ProposedChangeArtifactDefinition]:
    """This function assumes that definitions is a list of the edges

    The edge should be of type CoreArtifactDefinition from the query
    * GATHER_ARTIFACT_DEFINITIONS
    """

    parsed = []
    for definition in definitions:
        artifact_definition = ProposedChangeArtifactDefinition(
            definition_id=definition["node"]["id"],
            definition_name=definition["node"]["name"]["value"],
            artifact_name=definition["node"]["artifact_name"]["value"],
            content_type=definition["node"]["content_type"]["value"],
            timeout=definition["node"]["transformation"]["node"]["timeout"]["value"],
            query_name=definition["node"]["transformation"]["node"]["query"]["node"]["name"]["value"],
            query_models=definition["node"]["transformation"]["node"]["query"]["node"]["models"]["value"] or [],
            repository_id=definition["node"]["transformation"]["node"]["repository"]["node"]["id"],
            transform_kind=definition["node"]["transformation"]["node"]["__typename"],
        )
        if artifact_definition.transform_kind == InfrahubKind.TRANSFORMJINJA2:
            artifact_definition.template_path = definition["node"]["transformation"]["node"]["template_path"]["value"]
        elif artifact_definition.transform_kind == InfrahubKind.TRANSFORMPYTHON:
            artifact_definition.class_name = definition["node"]["transformation"]["node"]["class_name"]["value"]
            artifact_definition.file_path = definition["node"]["transformation"]["node"]["file_path"]["value"]
            artifact_definition.convert_query_response = definition["node"]["transformation"]["node"][
                "convert_query_response"
            ]["value"]

        parsed.append(artifact_definition)

    return parsed


async def _get_proposed_change_repositories(
    model: RequestProposedChangePipeline, client: InfrahubClient
) -> list[ProposedChangeRepository]:
    destination_all = await client.execute_graphql(
        query=DESTINATION_ALLREPOSITORIES, branch_name=model.destination_branch
    )
    source_managed = await client.execute_graphql(query=SOURCE_REPOSITORIES, branch_name=model.source_branch)
    source_readonly = await client.execute_graphql(query=SOURCE_READONLY_REPOSITORIES, branch_name=model.source_branch)

    destination_all = destination_all[InfrahubKind.GENERICREPOSITORY]["edges"]
    source_all = (
        source_managed[InfrahubKind.REPOSITORY]["edges"] + source_readonly[InfrahubKind.READONLYREPOSITORY]["edges"]
    )

    return _parse_proposed_change_repositories(model=model, source=source_all, destination=destination_all)


@task(
    name="proposed-change-validate-repository-conflicts",
    task_run_name="Validate conflicts on repository",
    cache_policy=NONE,
)  # type: ignore[arg-type]
async def _validate_repository_merge_conflicts(
    repositories: list[ProposedChangeRepository], client: InfrahubClient
) -> bool:
    log = get_run_logger()

    conflicts = False
    for repo in repositories:
        if repo.has_diff and not repo.is_staging:
            git_repo = await InfrahubRepository.init(id=repo.repository_id, name=repo.repository_name, client=client)
            async with lock.registry.get(name=repo.repository_name, namespace="repository"):
                repo.conflicts = await git_repo.get_conflicts(
                    source_branch=repo.source_branch, dest_branch=repo.destination_branch
                )
                if repo.conflicts:
                    log.info(f"{len(repo.conflicts)} conflict(s) identified on {repo.repository_name}")
                    conflicts = True
                else:
                    log.info(f"no conflict identified for {repo.repository_name}")

    return conflicts


async def _gather_repository_repository_diffs(
    repositories: list[ProposedChangeRepository], client: InfrahubClient
) -> None:
    for repo in repositories:
        if repo.has_diff and repo.source_commit and repo.destination_commit:
            # TODO we need to find a way to return all files in the repo if the repo is new
            git_repo = await InfrahubRepository.init(id=repo.repository_id, name=repo.repository_name, client=client)

            files_changed: list[str] = []
            files_added: list[str] = []
            files_removed: list[str] = []

            if repo.destination_branch:
                files_changed, files_added, files_removed = await git_repo.calculate_diff_between_commits(
                    first_commit=repo.source_commit, second_commit=repo.destination_commit
                )
            else:
                files_added = await git_repo.list_all_files(commit=repo.source_commit)

            repo.files_removed = files_removed
            repo.files_added = files_added
            repo.files_changed = files_changed


async def _populate_subscribers(
    branch_diff: ProposedChangeBranchDiff, diff_summary: list[NodeDiff], branch: str, client: InfrahubClient
) -> None:
    result = await client.execute_graphql(
        query=GATHER_GRAPHQL_QUERY_SUBSCRIBERS,
        branch_name=branch,
        variables={"members": get_modified_node_ids(diff_summary=diff_summary, branch=branch)},
    )

    for group in result[InfrahubKind.GRAPHQLQUERYGROUP]["edges"]:
        for subscriber in group["node"]["subscribers"]["edges"]:
            branch_diff.subscribers.append(
                ProposedChangeSubscriber(subscriber_id=subscriber["node"]["id"], kind=subscriber["node"]["__typename"])
            )



from ...generic_schema import GenericSchema

lineage_owner = GenericSchema(
    name="Owner",
    namespace="Lineage",
    description="Any Entities that is responsible for some data.",
    label="Owner",
    include_in_menu=False,
    documentation="/topics/metadata",
)

lineage_source = GenericSchema(
    name="Source",
    namespace="Lineage",
    description="Any Entities that stores or produces data.",
    label="Source",
    include_in_menu=False,
    documentation="/topics/metadata",
)



from infrahub.core.account import GlobalPermission
from infrahub.core.constants import GLOBAL_BRANCH_NAME, GlobalPermissions, PermissionDecision
from infrahub.core.registry import registry


def define_global_permission_from_branch(permission: GlobalPermissions, branch_name: str) -> GlobalPermission:
    if branch_name in (GLOBAL_BRANCH_NAME, registry.default_branch):
        decision = PermissionDecision.ALLOW_DEFAULT
    else:
        decision = PermissionDecision.ALLOW_OTHER

    return GlobalPermission(action=permission.value, decision=decision.value)



from infrahub.core.constants import (
    BranchSupportType,
    InfrahubKind,
    NumberPoolType,
)
from infrahub.core.constants import RelationshipCardinality as Cardinality
from infrahub.core.constants import RelationshipKind as RelKind

from ...attribute_schema import AttributeSchema as Attr
from ...generic_schema import GenericSchema
from ...node_schema import NodeSchema
from ...relationship_schema import (
    RelationshipSchema as Rel,
)

core_resource_pool = GenericSchema(
    name="ResourcePool",
    namespace="Core",
    label="Resource Pool",
    description="The resource manager contains pools of resources to allow for automatic assignments.",
    include_in_menu=False,
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    human_friendly_id=["name__value"],
    icon="mdi:view-grid-outline",
    branch=BranchSupportType.AGNOSTIC,
    uniqueness_constraints=[["name__value"]],
    generate_profile=False,
    attributes=[
        Attr(name="name", kind="Text", order_weight=1000, unique=True),
        Attr(name="description", kind="Text", optional=True, order_weight=2000),
    ],
)

core_weighted_pool_resource = GenericSchema(
    name="WeightedPoolResource",
    namespace="Core",
    label="Weighted Pool Resource",
    description="Resource to be used in a pool, its weight is used to determine its priority on allocation.",
    include_in_menu=False,
    branch=BranchSupportType.AWARE,
    generate_profile=False,
    attributes=[
        Attr(
            name="allocation_weight",
            label="Weight",
            description="Weight determines allocation priority, resources with higher values are selected first.",
            kind="Number",
            optional=True,
            order_weight=10000,
        )
    ],
)

core_ip_prefix_pool = NodeSchema(
    name="IPPrefixPool",
    namespace="Core",
    description="A pool of IP prefix resources",
    label="IP Prefix Pool",
    include_in_menu=False,
    branch=BranchSupportType.AGNOSTIC,
    generate_profile=False,
    inherit_from=[InfrahubKind.RESOURCEPOOL, InfrahubKind.LINEAGESOURCE],
    attributes=[
        Attr(
            name="default_prefix_length",
            kind="Number",
            description="The default prefix length as an integer for prefixes allocated from this pool.",
            optional=True,
            order_weight=5000,
        ),
        Attr(
            name="default_member_type",
            kind="Text",
            enum=["prefix", "address"],
            default_value="prefix",
            optional=True,
            order_weight=3000,
        ),
        Attr(name="default_prefix_type", kind="Text", optional=True, order_weight=4000),
    ],
    relationships=[
        Rel(
            name="resources",
            peer=InfrahubKind.IPPREFIX,
            kind=RelKind.ATTRIBUTE,
            identifier="prefixpool__resource",
            cardinality=Cardinality.MANY,
            branch=BranchSupportType.AGNOSTIC,
            optional=False,
            order_weight=6000,
        ),
        Rel(
            name="ip_namespace",
            peer=InfrahubKind.IPNAMESPACE,
            kind=RelKind.ATTRIBUTE,
            identifier="prefixpool__ipnamespace",
            cardinality=Cardinality.ONE,
            branch=BranchSupportType.AGNOSTIC,
            optional=False,
            order_weight=7000,
        ),
    ],
)

core_ip_address_pool = NodeSchema(
    name="IPAddressPool",
    namespace="Core",
    description="A pool of IP address resources",
    label="IP Address Pool",
    include_in_menu=False,
    branch=BranchSupportType.AGNOSTIC,
    generate_profile=False,
    inherit_from=[InfrahubKind.RESOURCEPOOL, InfrahubKind.LINEAGESOURCE],
    attributes=[
        Attr(
            name="default_address_type",
            kind="Text",
            optional=False,
            description="The object type to create when reserving a resource in the pool",
            order_weight=3000,
        ),
        Attr(
            name="default_prefix_length",
            kind="Number",
            optional=True,
            description="The default prefix length as an integer for addresses allocated from this pool.",
            order_weight=4000,
        ),
    ],
    relationships=[
        Rel(
            name="resources",
            peer=InfrahubKind.IPPREFIX,
            kind=RelKind.ATTRIBUTE,
            identifier="ipaddresspool__resource",
            cardinality=Cardinality.MANY,
            branch=BranchSupportType.AGNOSTIC,
            optional=False,
            order_weight=5000,
        ),
        Rel(
            name="ip_namespace",
            peer=InfrahubKind.IPNAMESPACE,
            kind=RelKind.ATTRIBUTE,
            identifier="ipaddresspool__ipnamespace",
            cardinality=Cardinality.ONE,
            branch=BranchSupportType.AGNOSTIC,
            optional=False,
            order_weight=6000,
        ),
    ],
)

core_number_pool = NodeSchema(
    name="NumberPool",
    namespace="Core",
    description="A pool of number resources",
    label="Number Pool",
    include_in_menu=False,
    branch=BranchSupportType.AGNOSTIC,
    generate_profile=False,
    inherit_from=[InfrahubKind.RESOURCEPOOL, InfrahubKind.LINEAGESOURCE],
    attributes=[
        Attr(
            name="node",
            kind="Text",
            optional=False,
            description="The model of the object that requires integers to be allocated",
            order_weight=3000,
        ),
        Attr(
            name="node_attribute",
            kind="Text",
            description="The attribute of the selected model",
            optional=False,
            order_weight=4000,
        ),
        Attr(
            name="start_range",
            kind="Number",
            optional=False,
            description="The start range for the pool",
            order_weight=5000,
        ),
        Attr(
            name="end_range", kind="Number", optional=False, description="The end range for the pool", order_weight=6000
        ),
        Attr(
            name="pool_type",
            kind="Text",
            description="Defines how this number pool was created",
            default_value=NumberPoolType.USER.value,
            enum=NumberPoolType.available_types(),
            read_only=True,
        ),
    ],
)



from __future__ import annotations

import ipaddress
from typing import TYPE_CHECKING, Any, Callable, get_args

from ..protocols_base import CoreNodeBase
from ..uuidt import UUIDT
from .constants import IP_TYPES, PROPERTIES_FLAG, PROPERTIES_OBJECT, SAFE_VALUE
from .property import NodeProperty

if TYPE_CHECKING:
    from ..schema import AttributeSchemaAPI


class Attribute:
    """Represents an attribute of a Node, including its schema, value, and properties."""

    def __init__(self, name: str, schema: AttributeSchemaAPI, data: Any | dict):
        """
        Args:
            name (str): The name of the attribute.
            schema (AttributeSchema): The schema defining the attribute.
            data (Union[Any, dict]): The data for the attribute, either in raw form or as a dictionary.
        """
        self.name = name
        self._schema = schema

        if not isinstance(data, dict) or "value" not in data.keys():
            data = {"value": data}

        self._properties_flag = PROPERTIES_FLAG
        self._properties_object = PROPERTIES_OBJECT
        self._properties = self._properties_flag + self._properties_object

        self._read_only = ["updated_at", "is_inherited"]

        self.id: str | None = data.get("id", None)

        self._value: Any | None = data.get("value", None)
        self.value_has_been_mutated = False
        self.is_default: bool | None = data.get("is_default", None)
        self.is_from_profile: bool | None = data.get("is_from_profile", None)

        if self._value:
            value_mapper: dict[str, Callable] = {
                "IPHost": ipaddress.ip_interface,
                "IPNetwork": ipaddress.ip_network,
            }
            mapper = value_mapper.get(schema.kind, lambda value: value)
            self._value = mapper(data.get("value"))

        self.is_inherited: bool | None = data.get("is_inherited", None)
        self.updated_at: str | None = data.get("updated_at", None)

        self.is_visible: bool | None = data.get("is_visible", None)
        self.is_protected: bool | None = data.get("is_protected", None)

        self.source: NodeProperty | None = None
        self.owner: NodeProperty | None = None

        for prop_name in self._properties_object:
            if data.get(prop_name):
                setattr(self, prop_name, NodeProperty(data=data.get(prop_name)))  # type: ignore[arg-type]

    @property
    def value(self) -> Any:
        return self._value

    @value.setter
    def value(self, value: Any) -> None:
        self._value = value
        self.value_has_been_mutated = True

    def _generate_input_data(self) -> dict | None:
        data: dict[str, Any] = {}
        variables: dict[str, Any] = {}

        if self.value is None:
            return data

        if isinstance(self.value, str):
            if SAFE_VALUE.match(self.value):
                data["value"] = self.value
            else:
                var_name = f"value_{UUIDT.new().hex}"
                variables[var_name] = self.value
                data["value"] = f"${var_name}"
        elif isinstance(self.value, get_args(IP_TYPES)):
            data["value"] = self.value.with_prefixlen
        elif isinstance(self.value, CoreNodeBase) and self.value.is_resource_pool():
            data["from_pool"] = {"id": self.value.id}
        else:
            data["value"] = self.value

        for prop_name in self._properties_flag:
            if getattr(self, prop_name) is not None:
                data[prop_name] = getattr(self, prop_name)

        for prop_name in self._properties_object:
            if getattr(self, prop_name) is not None:
                data[prop_name] = getattr(self, prop_name)._generate_input_data()

        return {"data": data, "variables": variables}

    def _generate_query_data(self, property: bool = False) -> dict | None:
        data: dict[str, Any] = {"value": None}

        if property:
            data.update({"is_default": None, "is_from_profile": None})

            for prop_name in self._properties_flag:
                data[prop_name] = None
            for prop_name in self._properties_object:
                data[prop_name] = {"id": None, "display_label": None, "__typename": None}

        return data

    def _generate_mutation_query(self) -> dict[str, Any]:
        if isinstance(self.value, CoreNodeBase) and self.value.is_resource_pool():
            # If it points to a pool, ask for the value of the pool allocated resource
            return {self.name: {"value": None}}
        return {}



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from starlette.background import BackgroundTasks

from infrahub.context import InfrahubContext
from infrahub.core import registry
from infrahub.core.timestamp import Timestamp
from infrahub.exceptions import InitializationError
from infrahub.graphql.resolvers.many_relationship import ManyRelationshipResolver
from infrahub.graphql.resolvers.single_relationship import SingleRelationshipResolver
from infrahub.permissions import PermissionManager

from .manager import GraphQLSchemaManager

if TYPE_CHECKING:
    from graphql import GraphQLSchema
    from starlette.requests import HTTPConnection

    from infrahub.auth import AccountSession
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase
    from infrahub.services import InfrahubServices


@dataclass
class GraphqlParams:
    schema: GraphQLSchema
    context: GraphqlContext


@dataclass
class GraphqlContext:
    db: InfrahubDatabase
    branch: Branch
    types: dict
    single_relationship_resolver: SingleRelationshipResolver
    many_relationship_resolver: ManyRelationshipResolver
    service: InfrahubServices | None = None
    at: Timestamp | None = None
    related_node_ids: set | None = None
    account_session: AccountSession | None = None
    permissions: PermissionManager | None = None
    background: BackgroundTasks | None = None
    request: HTTPConnection | None = None

    @property
    def active_account_session(self) -> AccountSession:
        """Return an account session or raise an error

        Eventualy this property should be removed, that can be done after self.account_session is no longer optional
        """
        if self.account_session:
            return self.account_session
        raise InitializationError("GraphQLContext doesn't contain an account_session")

    @property
    def active_permissions(self) -> PermissionManager:
        """Return a permission manager or raise an error

        This property should be removed, once self.account_session is no longer optional which will imply self.permissions will no longer be optional
        as well.
        """
        if self.permissions:
            return self.permissions
        raise InitializationError("GraphQLContext doesn't contain permissions")

    @property
    def active_service(self) -> InfrahubServices:
        if self.service:
            return self.service
        raise InitializationError("GraphQLContext doesn't contain a service")

    def get_context(self) -> InfrahubContext:
        return InfrahubContext.init(branch=self.branch, account=self.active_account_session)


async def prepare_graphql_params(
    db: InfrahubDatabase,
    branch: Branch | str,
    at: Timestamp | str | None = None,
    account_session: AccountSession | None = None,
    request: HTTPConnection | None = None,
    service: InfrahubServices | None = None,
    include_query: bool = True,
    include_mutation: bool = True,
    include_subscription: bool = True,
    include_types: bool = True,
) -> GraphqlParams:
    branch = registry.get_branch_from_registry(branch=branch)
    schema_branch = registry.schema.get_schema_branch(name=branch.name)
    gqlm = GraphQLSchemaManager.get_manager_for_branch(branch=branch, schema_branch=schema_branch)
    gql_schema = gqlm.get_graphql_schema(
        include_query=include_query,
        include_mutation=include_mutation,
        include_subscription=include_subscription,
        include_types=include_types,
    )

    if request and not service:
        service = request.app.state.service

    permissions: PermissionManager | None = None
    if account_session:
        permissions = PermissionManager(account_session=account_session)
        await permissions.load_permissions(db=db, branch=branch)

    return GraphqlParams(
        schema=gql_schema,
        context=GraphqlContext(
            db=db,
            branch=branch,
            single_relationship_resolver=SingleRelationshipResolver(),
            many_relationship_resolver=ManyRelationshipResolver(),
            at=Timestamp(at),
            types=gqlm.get_graphql_types(),
            related_node_ids=set(),
            background=BackgroundTasks(),
            request=request,
            service=service,
            account_session=account_session,
            permissions=permissions,
        ),
    )



"""Config Class."""

from __future__ import annotations

from pathlib import Path

import toml
import typer
from pydantic import Field, ValidationError, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

DEFAULT_CONFIG_FILE = "infrahubctl.toml"
ENVVAR_CONFIG_FILE = "INFRAHUBCTL_CONFIG"
INFRAHUB_REPO_CONFIG_FILE = ".infrahub.yml"


class Settings(BaseSettings):
    """Main Settings Class for the project."""

    model_config = SettingsConfigDict(env_prefix="INFRAHUB_", populate_by_name=True, extra="allow")
    server_address: str = Field(default="http://localhost:8000", validation_alias="infrahub_address")
    api_token: str | None = Field(default=None)
    default_branch: str = Field(default="main")

    @field_validator("server_address")
    @classmethod
    def cleanup_server_address(cls, v: str) -> str:
        return v.rstrip("/")


class ConfiguredSettings:
    def __init__(self) -> None:
        self._settings: Settings | None = None

    @property
    def active(self) -> Settings:
        if self._settings:
            return self._settings

        print("Configuration not properly loaded")
        raise typer.Abort()

    def load(self, config_file: str | Path = "infrahubctl.toml", config_data: dict | None = None) -> None:
        """Load configuration.

        Configuration is loaded from a config file in toml format that contains the settings,
        or from a dictionary of those settings passed in as "config_data"
        """

        if self._settings:
            return

        if config_data:
            self._settings = Settings(**config_data)
            return

        if not isinstance(config_file, Path):
            config_file = Path(config_file)

        if config_file.is_file():
            config_string = config_file.read_text(encoding="utf-8")
            config_tmp = toml.loads(config_string)

            self._settings = Settings(**config_tmp)
            return

        self._settings = Settings()

    def load_and_exit(self, config_file: str | Path = "infrahubctl.toml", config_data: dict | None = None) -> None:
        """Calls load, but wraps it in a try except block.

        This is done to handle a ValidationErorr which is raised when settings are specified but invalid.
        In such cases, a message is printed to the screen indicating the settings which don't pass validation.

        Args:
            config_file_name (str, optional): [description]. Defaults to "pyprojectctl.toml".
            config_data (dict, optional): [description]. Defaults to None.
        """

        try:
            self.load(config_file=config_file, config_data=config_data)
        except ValidationError as exc:
            print(f"Configuration not valid, found {len(exc.errors())} error(s)")
            for error in exc.errors():
                loc_str = [str(item) for item in error["loc"]]
                print(f"  {'/'.join(loc_str)} | {error['msg']} ({error['type']})")
            raise typer.Abort()


SETTINGS = ConfiguredSettings()



from infrahub.core.validators.aggregated_checker import AggregatedConstraintChecker
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext

from .attribute_choices import SchemaAttributeChoicesConstraintDependency
from .attribute_enum import SchemaAttributeEnumConstraintDependency
from .attribute_kind import SchemaAttributeKindConstraintDependency
from .attribute_length import SchemaAttributLengthConstraintDependency
from .attribute_optional import SchemaAttributeOptionalConstraintDependency
from .attribute_regex import SchemaAttributeRegexConstraintDependency
from .attribute_uniqueness import SchemaAttributeUniqueConstraintDependency
from .generate_profile import SchemaGenerateProfileConstraintDependency
from .inherit_from import SchemaInheritFromConstraintDependency
from .node_attribute import SchemaNodeAttributeAddConstraintDependency
from .node_relationship import SchemaNodeRelationshipAddConstraintDependency
from .relationship_count import SchemaRelationshipCountConstraintDependency
from .relationship_optional import SchemaRelationshipOptionalConstraintDependency
from .relationship_peer import SchemaRelationshipPeerParentConstraintDependency
from .uniqueness import SchemaUniquenessConstraintDependency


class AggregatedSchemaConstraintsDependency(DependencyBuilder[AggregatedConstraintChecker]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> AggregatedConstraintChecker:
        return AggregatedConstraintChecker(
            constraints=[
                SchemaUniquenessConstraintDependency.build(context=context),
                SchemaGenerateProfileConstraintDependency.build(context=context),
                SchemaInheritFromConstraintDependency.build(context=context),
                SchemaRelationshipOptionalConstraintDependency.build(context=context),
                SchemaRelationshipCountConstraintDependency.build(context=context),
                SchemaAttributeRegexConstraintDependency.build(context=context),
                SchemaAttributeUniqueConstraintDependency.build(context=context),
                SchemaAttributeOptionalConstraintDependency.build(context=context),
                SchemaAttributeChoicesConstraintDependency.build(context=context),
                SchemaAttributeEnumConstraintDependency.build(context=context),
                SchemaAttributLengthConstraintDependency.build(context=context),
                SchemaAttributeKindConstraintDependency.build(context=context),
                SchemaNodeAttributeAddConstraintDependency.build(context=context),
                SchemaNodeRelationshipAddConstraintDependency.build(context=context),
                SchemaRelationshipPeerParentConstraintDependency.build(context=context),
            ],
            db=context.db,
            branch=context.branch,
        )



from __future__ import annotations

import base64
import hashlib
import json
import uuid
from itertools import groupby
from pathlib import Path
from typing import TYPE_CHECKING, Any
from uuid import UUID, uuid4

import httpx
import ujson
from graphql import (
    FieldNode,
    InlineFragmentNode,
    SelectionSetNode,
)

from infrahub_sdk.repository import GitRepoManager

from .exceptions import FileNotValidError, JsonDecodeError, TimestampFormatError
from .timestamp import Timestamp

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo
    from whenever import TimeDelta


def generate_short_id() -> str:
    """Generate a short unique ID"""
    return base64.urlsafe_b64encode(uuid.uuid4().bytes).rstrip(b"=").decode("ascii").lower()


def base36encode(number: int) -> str:
    if not isinstance(number, (int)):
        raise TypeError("number must be an integer")
    is_negative = number < 0
    number = abs(number)

    alphabet = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    base36 = ""

    while number:
        number, i = divmod(number, 36)
        base36 = alphabet[i] + base36
    if is_negative:
        base36 = "-" + base36

    return base36 or alphabet[0]


def base36decode(data: str) -> int:
    return int(data, 36)


def base16decode(data: str) -> int:
    return int(data, 16)


def base16encode(number: int) -> str:
    if not isinstance(number, (int)):
        raise TypeError("number must be an integer")
    is_negative = number < 0
    number = abs(number)

    alphabet = "0123456789abcdef"
    base16 = ""

    while number:
        number, i = divmod(number, 16)
        base16 = alphabet[i] + base16
    if is_negative:
        base16 = "-" + base16

    return base16 or alphabet[0]


def get_fixtures_dir() -> Path:
    """Get the directory which stores fixtures that are common to multiple unit/integration tests."""
    here = Path(__file__).resolve().parent
    return here.parent / "tests" / "fixtures"


def is_valid_uuid(value: Any) -> bool:
    """Check if the input is a valid UUID."""
    try:
        UUID(str(value))
        return True
    except ValueError:
        return False


def decode_json(response: httpx.Response) -> dict:
    try:
        return response.json()
    except json.decoder.JSONDecodeError as exc:
        raise JsonDecodeError(content=response.text, url=response.url) from exc


def generate_uuid() -> str:
    return str(uuid4())


def duplicates(input_list: list) -> list:
    """Identify and return all the duplicates in a list."""

    dups = []

    clean_input_list = [item for item in input_list or [] if item is not None]
    for x, y in groupby(sorted(clean_input_list)):
        #  list(y) returns all the occurences of item x
        if len(list(y)) > 1:
            dups.append(x)

    return dups


def intersection(list1: list[Any], list2: list[Any]) -> list:
    """Calculate the intersection between 2 lists."""
    return list(set(list1) & set(list2))


def compare_lists(list1: list[Any], list2: list[Any]) -> tuple[list[Any], list[Any], list[Any]]:
    """Compare 2 lists and return :
    - the intersection of both
    - the item present only in list1
    - the item present only in list2
    """

    in_both = intersection(list1=list1, list2=list2)
    in_list_1 = list(set(list1) - set(in_both))
    in_list_2 = list(set(list2) - set(in_both))

    return sorted(in_both), sorted(in_list_1), sorted(in_list_2)


def deep_merge_dict(dicta: dict, dictb: dict, path: list | None = None) -> dict:
    """Deep Merge Dictionary B into Dictionary A.
    Code is inspired by https://stackoverflow.com/a/7205107
    """
    if path is None:
        path = []
    for key in dictb:
        if key in dicta:
            if isinstance(dicta[key], dict) and isinstance(dictb[key], dict):
                deep_merge_dict(dicta[key], dictb[key], path + [str(key)])
            elif isinstance(dicta[key], list) and isinstance(dictb[key], list):
                # Merge lists
                # Cannot use compare_list because list of dicts won't work (dict not hashable)
                dicta[key] = [i for i in dicta[key] if i not in dictb[key]] + dictb[key]
            elif dicta[key] == dictb[key]:
                continue
            else:
                raise ValueError("Conflict at %s" % ".".join(path + [str(key)]))
        else:
            dicta[key] = dictb[key]
    return dicta


def str_to_bool(value: str) -> bool:
    """Convert a String to a Boolean"""

    if isinstance(value, bool):
        return value

    if isinstance(value, int) and value in [0, 1]:
        return bool(value)

    if not isinstance(value, str):
        raise TypeError(f"{value} must be a string")

    MAP = {
        "y": True,
        "yes": True,
        "t": True,
        "true": True,
        "on": True,
        "1": True,
        "n": False,
        "no": False,
        "f": False,
        "false": False,
        "off": False,
        "0": False,
    }
    try:
        return MAP[value.lower()]
    except KeyError as exc:
        raise ValueError(f"{value} can not be converted into a boolean") from exc


def generate_request_filename(request: httpx.Request) -> str:
    """Return a filename for a request sent to the Infrahub API

    This function is used when recording and playing back requests, as Infrahub is using a GraphQL
    API it's not possible to rely on the URL endpoint alone to separate one request from another,
    for this reason a hash of the payload is included in a filename.
    """
    formatted = (
        str(request.url).replace(":", "_").replace("//", "").replace("/", "__").replace("?", "_q_").replace("&", "_a_")
    )
    filename = f"{request.method}_{formatted}"
    if request.content:
        content_hash = hashlib.sha224(request.content)
        filename += f"_{content_hash.hexdigest()}"

    return filename.lower()


def is_valid_url(url: str) -> bool:
    if not isinstance(url, str):
        return False
    if "://" not in url and not url.startswith("/"):
        return False
    if "://" not in url:
        url = "http://localhost" + url

    try:
        parsed = httpx.URL(url)
        return all([parsed.scheme, parsed.netloc])
    except TypeError:
        return False


def get_branch(branch: str | None = None, directory: str | Path = ".") -> str:
    """If branch isn't provide, return the name of the local Git branch."""
    if branch:
        return branch

    repo = GitRepoManager(directory)
    return str(repo.active_branch)


def dict_hash(dictionary: dict[str, Any]) -> str:
    """MD5 hash of a dictionary."""
    # We need to sort arguments so {'a': 1, 'b': 2} is
    # the same as {'b': 2, 'a': 1}
    encoded = ujson.dumps(dictionary, sort_keys=True).encode()
    dhash = hashlib.md5(encoded, usedforsecurity=False)
    return dhash.hexdigest()


def calculate_dict_depth(data: dict, level: int = 1) -> int:
    """Calculate the depth of a nested Dictionary recursively."""
    if not isinstance(data, dict) or not data:
        return level
    return max(calculate_dict_depth(data=data[key], level=level + 1) for key in data)


def calculate_dict_height(data: dict, cnt: int = 0) -> int:
    """Calculate the number of fields (height) in a nested Dictionary recursively."""
    for key in data:
        if isinstance(data[key], dict):
            cnt = calculate_dict_height(data=data[key], cnt=cnt + 1)
        else:
            cnt += 1
    return cnt


async def extract_fields(selection_set: SelectionSetNode | None) -> dict[str, dict] | None:
    """This function extract all the requested fields in a tree of Dict from a SelectionSetNode

    The goal of this function is to limit the fields that we need to query from the backend.

    Currently the function support Fields and InlineFragments but in a combined tree where the fragments are merged together
    This implementation may seam counter intuitive but in the current implementation
    it's better to have slightly more information at time passed to the query manager.

    In the future we'll probably need to redesign how we read GraphQL queries to generate better Database query.
    """

    if not selection_set:
        return None

    fields = {}
    for node in selection_set.selections:
        sub_selection_set = getattr(node, "selection_set", None)
        if isinstance(node, FieldNode):
            value = await extract_fields(sub_selection_set)
            if node.name.value not in fields:
                fields[node.name.value] = value
            elif isinstance(fields[node.name.value], dict) and isinstance(value, dict):
                fields[node.name.value].update(value)

        elif isinstance(node, InlineFragmentNode):
            for sub_node in node.selection_set.selections:
                sub_sub_selection_set = getattr(sub_node, "selection_set", None)
                value = await extract_fields(sub_sub_selection_set)
                if sub_node.name.value not in fields:
                    fields[sub_node.name.value] = await extract_fields(sub_sub_selection_set)
                elif isinstance(fields[sub_node.name.value], dict) and isinstance(value, dict):
                    fields[sub_node.name.value].update(value)

    return fields


async def extract_fields_first_node(info: GraphQLResolveInfo) -> dict[str, dict]:
    fields = None
    if info.field_nodes:
        fields = await extract_fields(info.field_nodes[0].selection_set)

    return fields or {}


def write_to_file(path: Path, value: Any) -> bool:
    """Write a given value into a file and return if the operation was successful.

    If the file does not exist, the function will attempt to create it."""
    if not path.exists():
        path.touch()

    if path.is_dir():
        raise FileExistsError(f"{path} is a directory")

    to_write = str(value)
    written = path.write_text(to_write)

    return written is not None


def read_file(file_path: Path) -> str:
    if not file_path.is_file():
        raise FileNotValidError(name=str(file_path.name), message=f"{file_path.name}: not found at {file_path.parent}")
    try:
        with Path.open(file_path, encoding="utf-8") as fobj:
            return fobj.read()
    except UnicodeDecodeError as exc:
        raise FileNotValidError(
            name=str(file_path.name), message=f"Unable to read {file_path.name} with utf-8 encoding"
        ) from exc


def get_user_permissions(data: list[dict]) -> dict:
    groups = {}
    for group in data:
        group_name = group["node"]["display_label"]
        permissions = []

        roles = group["node"].get("roles", {}).get("edges", [])
        for role in roles:
            role_permissions = role["node"].get("permissions", {}).get("edges", [])
            for permission in role_permissions:
                permissions.append(permission["node"]["identifier"]["value"])

        groups[group_name] = permissions

    return groups


def calculate_time_diff(value: str) -> str | None:
    """Calculate the time in human format between a timedate in string format and now."""
    try:
        time_value = Timestamp(value)
    except TimestampFormatError:
        return None

    delta: TimeDelta = Timestamp().get_obj().difference(time_value.get_obj())
    (hrs, mins, secs, nanos) = delta.in_hrs_mins_secs_nanos()

    if nanos and nanos > 500_000_000:
        secs += 1

    if hrs and hrs < 24 and mins:
        return f"{hrs}h {mins}m and {secs}s ago"
    if hrs and hrs > 24:
        remaining_hrs = hrs % 24
        days = int((hrs - remaining_hrs) / 24)
        return f"{days}d and {remaining_hrs}h ago"
    if hrs == 0 and mins and secs:
        return f"{mins}m and {secs}s ago"
    if hrs == 0 and mins == 0 and secs:
        return f"{secs}s ago"
    return "now"



from __future__ import annotations

from collections.abc import Iterable, Mapping, Sequence
from itertools import chain
from typing import Any


class DependencyCycleExistsError(Exception):
    def __init__(self, cycles: Iterable[Sequence[str]], *args: tuple[Any]) -> None:
        self.cycles = cycles
        super().__init__(*args)

    def get_cycle_strings(self) -> list[str]:
        return [" --> ".join([str(node) for node in cycle]) for cycle in self.cycles]

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.get_cycle_strings()})"


def topological_sort(dependency_dict: Mapping[str, Iterable[str]]) -> list[set[str]]:
    if not dependency_dict:
        return []

    missing_dependent_keys = set(chain(*dependency_dict.values())) - set(dependency_dict.keys())

    dependency_dict_to_sort = {k: set(v) for k, v in dependency_dict.items()}
    dependency_dict_to_sort.update({missing_key: set() for missing_key in missing_dependent_keys})

    ordered = []
    while len(dependency_dict_to_sort) > 0:
        nondependant_nodes = {key for key, dependencies in dependency_dict_to_sort.items() if len(dependencies) == 0}
        dependency_dict_to_sort = {
            k: v - nondependant_nodes for k, v in dependency_dict_to_sort.items() if k not in nondependant_nodes
        }
        ordered.append(nondependant_nodes)

        if len(nondependant_nodes) == 0 and len(dependency_dict_to_sort) != 0:
            cycles = get_cycles(dependency_dict_to_sort)
            raise DependencyCycleExistsError(cycles=cycles)
    return ordered


def get_cycles(dependency_dict: Mapping[str, Iterable[str]]) -> list[list[str]]:
    if not dependency_dict:
        return []

    dict_to_check = {**dependency_dict}
    cycles = []

    while dict_to_check:
        start_path = list(dict_to_check.keys())[:1]
        cycles += _get_cycles(dependency_dict=dict_to_check, path=start_path)
    return cycles


def _get_cycles(dependency_dict: dict[str, Iterable[str]], path: list[str]) -> list[list[str]]:
    try:
        next_nodes = dependency_dict.pop(path[-1])
    except KeyError:
        return []
    cycles = []
    for next_node in next_nodes:
        if next_node in path:
            cycles.append(path[path.index(next_node) :] + [next_node])
        else:
            next_cycles = _get_cycles(dependency_dict, path + [next_node])
            if next_cycles:
                cycles += next_cycles
    return cycles



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Boolean, DateTime, Field, InputObjectType, Int, Interface, List, NonNull, ObjectType, String
from graphene.types.generic import GenericScalar

from infrahub import events

from .common import RelatedNode
from .enums import DiffAction

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo


class InfrahubMutatedAttribute(ObjectType):
    name = String(required=True)
    action = DiffAction(required=True)
    value = String(required=False)
    kind = String(required=True)
    value_previous = String(required=False)


class InfrahubMutatedRelationship(ObjectType):
    name = String(required=True)
    action = DiffAction(required=True)
    peer = Field(RelatedNode, required=True)


class EventNodeInterface(Interface):
    id = String(required=True, description="The ID of the event.")
    event = String(required=True, description="The name of the event.")
    branch = String(required=False, description="The branch where the event occurred.")
    account_id = String(required=False, description="The account ID that triggered the event.")
    occurred_at = DateTime(required=True, description="The timestamp when the event occurred.")
    level = Int(
        required=True,
        description="The level of the event 0 is a root level event, the child events will have 1 and grand children 2.",
    )
    primary_node = Field(
        RelatedNode, required=False, description="The primary Infrahub node this event is associated with."
    )
    related_nodes = List(
        NonNull(RelatedNode), required=True, description="Related Infrahub nodes this event is associated with."
    )
    has_children = Boolean(
        required=True, description="Indicates if the event is expected to have child events under it"
    )
    parent_id = String(required=False, description="The event ID of the direct parent to this event.")

    @classmethod
    def resolve_type(
        cls,
        instance: dict[str, Any],
        info: GraphQLResolveInfo,  # noqa: ARG003
    ) -> type[ObjectType]:
        if "event" in instance:
            return EVENT_TYPES.get(instance["event"], StandardEvent)
        return StandardEvent


class EventNodes(ObjectType):
    node = Field(EventNodeInterface)


class BranchEventTypeFilter(InputObjectType):
    branches = List(NonNull(String), required=True, description="Name of impacted branches")


class EventTypeFilter(InputObjectType):
    branch_merged = Field(
        BranchEventTypeFilter, required=False, description="Filters specific to infrahub.branch.merged events"
    )
    branch_rebased = Field(
        BranchEventTypeFilter, required=False, description="Filters specific to infrahub.branch.rebased events"
    )


# ---------------------------------------
# Branch events
# ---------------------------------------
class BranchCreatedEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    created_branch = String(required=True, description="The name of the branch that was created")
    payload = Field(GenericScalar, required=True)


class BranchMergedEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    source_branch = String(required=True, description="The name of the branch that was merged into the default branch")


class BranchRebasedEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    rebased_branch = String(
        required=True, description="The name of the branch that was rebased and aligned with the default branch"
    )
    payload = Field(GenericScalar, required=True)


class BranchDeletedEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    deleted_branch = String(required=True, description="The name of the branch that was deleted")
    payload = Field(GenericScalar, required=True)


# ---------------------------------------
# Proposed change events
# ---------------------------------------
class ProposedChangeReviewEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    reviewer_account_id = String(required=True, description="The ID of the user who reviewed the proposed change")
    reviewer_account_name = String(required=True, description="The name of the user who reviewed the proposed change")
    reviewer_decision = String(required=True, description="The decision made by the reviewer")
    payload = Field(GenericScalar, required=True)


class ProposedChangeReviewRevokedEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    reviewer_account_id = String(required=True, description="The ID of the user who reviewed the proposed change")
    reviewer_account_name = String(required=True, description="The name of the user who reviewed the proposed change")
    reviewer_former_decision = String(required=True, description="The decision made by the reviewer")
    payload = Field(GenericScalar, required=True)


class ProposedChangeReviewRequestedEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    requested_by_account_id = String(
        required=True, description="The ID of the user who requested the proposed change to be reviewed"
    )
    requested_by_account_name = String(
        required=True, description="The name of the user who requested the proposed change to be reviewed"
    )
    payload = Field(GenericScalar, required=True)


class ProposedChangeMergedEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    merged_by_account_id = String(required=True, description="The ID of the user who merged the proposed change")
    merged_by_account_name = String(required=True, description="The name of the user who merged the proposed change")
    payload = Field(GenericScalar, required=True)


class ProposedChangeThreadEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    payload = Field(GenericScalar, required=True)


# ---------------------------------------
# Node/Object events
# ---------------------------------------
class NodeMutatedEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    payload = Field(GenericScalar, required=True)
    attributes = Field(List(of_type=NonNull(InfrahubMutatedAttribute), required=True), required=True)
    relationships = Field(List(of_type=NonNull(InfrahubMutatedRelationship), required=True), required=True)


class ArtifactEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    checksum = String(required=True, description="The current checksum of the artifact")
    checksum_previous = String(required=False, description="The previous checksum of the artifact")
    storage_id = String(required=True, description="The current storage_id of the artifact")
    storage_id_previous = String(required=False, description="The previous storage_id of the artifact")
    artifact_definition_id = String(required=True, description="Artifact definition ID")


class GroupEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    members = List(NonNull(RelatedNode), required=True, description="Group members modified in this event")
    ancestors = List(NonNull(RelatedNode), required=True, description="Ancestor groups of this impacted group")


class StandardEvent(ObjectType):
    class Meta:
        interfaces = (EventNodeInterface,)

    payload = Field(GenericScalar, required=True)


EVENT_TYPES: dict[str, type[ObjectType]] = {
    events.ArtifactCreatedEvent.event_name: ArtifactEvent,
    events.ArtifactUpdatedEvent.event_name: ArtifactEvent,
    events.NodeCreatedEvent.event_name: NodeMutatedEvent,
    events.NodeUpdatedEvent.event_name: NodeMutatedEvent,
    events.NodeDeletedEvent.event_name: NodeMutatedEvent,
    events.BranchCreatedEvent.event_name: BranchCreatedEvent,
    events.BranchMergedEvent.event_name: BranchMergedEvent,
    events.BranchRebasedEvent.event_name: BranchRebasedEvent,
    events.BranchDeletedEvent.event_name: BranchDeletedEvent,
    events.GroupMemberAddedEvent.event_name: GroupEvent,
    events.GroupMemberRemovedEvent.event_name: GroupEvent,
    events.ProposedChangeApprovedEvent.event_name: ProposedChangeReviewEvent,
    events.ProposedChangeApprovalRevokedEvent.event_name: ProposedChangeReviewRevokedEvent,
    events.ProposedChangeRejectedEvent.event_name: ProposedChangeReviewEvent,
    events.ProposedChangeRejectionRevokedEvent.event_name: ProposedChangeReviewRevokedEvent,
    events.ProposedChangeReviewRequestedEvent.event_name: ProposedChangeReviewRequestedEvent,
    events.ProposedChangeMergedEvent.event_name: ProposedChangeMergedEvent,
    events.ProposedChangeThreadCreatedEvent.event_name: ProposedChangeThreadEvent,
    events.ProposedChangeThreadUpdatedEvent.event_name: ProposedChangeThreadEvent,
    "undefined": StandardEvent,
}



from infrahub_sdk.groups import group_add_subscriber
from infrahub_sdk.utils import dict_hash
from prefect import flow

from infrahub.core.constants import InfrahubKind
from infrahub.groups.models import RequestGraphQLQueryGroupUpdate
from infrahub.workers.dependencies import get_client
from infrahub.workflows.utils import add_tags


@flow(name="graphql-query-group-update", flow_run_name="Update GraphQLQuery Group '{model.query_name}'")
async def update_graphql_query_group(model: RequestGraphQLQueryGroupUpdate) -> None:
    """Create or Update a GraphQLQueryGroup."""

    client = get_client()

    # If there is only one subscriber, associate the task to it
    # If there are more than one, for now we can't associate all of them
    related_nodes = []
    if len(model.subscribers) == 1:
        related_nodes.append(model.subscribers[0])

    await add_tags(branches=[model.branch], nodes=related_nodes)

    params_hash = dict_hash(model.params)
    group_name = f"{model.query_name}__{params_hash}"
    group_label = f"Query {model.query_name} Hash({params_hash[:8]})"
    group = await client.create(
        kind=InfrahubKind.GRAPHQLQUERYGROUP,
        branch=model.branch,
        name=group_name,
        label=group_label,
        group_type="internal",
        query=model.query_id,
        parameters=model.params,
        members=model.related_node_ids,
    )
    await group.save(allow_upsert=True)

    if model.subscribers:
        await group_add_subscriber(client=client, group=group, subscribers=model.subscribers, branch=model.branch)



from __future__ import annotations

from pydantic import Field

from infrahub.core.constants import InfrahubKind
from infrahub.core.schema.basenode_schema import BaseNodeSchema


class TemplateSchema(BaseNodeSchema):
    inherit_from: list[str] = Field(
        default_factory=list, description="List of Generic Kind that this template is inheriting from"
    )

    @property
    def is_node_schema(self) -> bool:
        return False

    @property
    def is_generic_schema(self) -> bool:
        return False

    @property
    def is_profile_schema(self) -> bool:
        return False

    @property
    def is_template_schema(self) -> bool:
        return True

    @property
    def is_ip_prefix(self) -> bool:
        return False

    @property
    def is_ip_address(self) -> bool:
        return False

    def get_labels(self) -> list[str]:
        """Return the labels for this object, composed of the kind and the list of Generic this object is inheriting from."""

        labels: list[str] = [self.kind] + self.inherit_from
        if self.namespace not in ["Schema", "Internal"] and InfrahubKind.GENERICGROUP not in self.inherit_from:
            labels.append(InfrahubKind.OBJECTTEMPLATE)
        return labels



from typing import Any

from graphql import (
    FieldNode,
    FragmentSpreadNode,
    GraphQLResolveInfo,
    InlineFragmentNode,
    SelectionSetNode,
)


class GraphQLFieldExtractor:
    """Class to extract fields from a GraphQL selection set."""

    def __init__(self, info: GraphQLResolveInfo):
        self.info = info
        self.fragments = info.fragments

    def get_fields(self) -> dict[str, Any]:
        """Extract fields from the GraphQL selection set."""
        fields = self._extract_fields(selection_set=self.info.field_nodes[0].selection_set)
        return fields or {}

    def _extract_fields(self, selection_set: SelectionSetNode | None) -> dict[str, dict] | None:
        """This function extract all the requested fields in a tree of Dict from a SelectionSetNode

        The goal of this function is to limit the fields that we need to query from the backend.

        Currently the function support Fields and InlineFragments but in a combined tree where the fragments are merged together
        This implementation may seam counter intuitive but in the current implementation
        it's better to have slightly more information at time passed to the query manager.

        In the future we'll probably need to redesign how we read GraphQL queries to generate better Database query.
        """

        if not selection_set:
            return None

        fields: dict[str, dict | Any] = {}
        for node in selection_set.selections:
            sub_selection_set = getattr(node, "selection_set", None)
            if isinstance(node, FieldNode):
                value = self._extract_fields(sub_selection_set)
                if node.name.value not in fields:
                    fields[node.name.value] = value
                elif isinstance(fields[node.name.value], dict) and isinstance(value, dict):
                    fields[node.name.value].update(value)

            elif isinstance(node, InlineFragmentNode):
                for sub_node in node.selection_set.selections:
                    sub_sub_selection_set = getattr(sub_node, "selection_set", None)
                    value = self._extract_fields(sub_sub_selection_set)
                    sub_node_name = getattr(sub_node, "name", "")
                    sub_node_name_value = getattr(sub_node_name, "value", "")
                    if sub_node_name_value not in fields:
                        fields[sub_node_name_value] = self._extract_fields(sub_sub_selection_set)
                    elif isinstance(fields[sub_node_name_value], dict) and isinstance(value, dict):
                        fields[sub_node_name_value].update(value)
            elif isinstance(node, FragmentSpreadNode):
                if node.name.value in self.info.fragments:
                    if fragment_fields := self._extract_fields(self.info.fragments[node.name.value].selection_set):
                        fields.update(fragment_fields)

        return fields


def extract_graphql_fields(info: GraphQLResolveInfo) -> dict[str, Any]:
    graphql_extractor = GraphQLFieldExtractor(info=info)
    return graphql_extractor.get_fields()
